/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Gee;
using Diorite.Logger;
using Diorite.System;

namespace Diorite{
	private const string OPTION_TRUE = "yes";
	private const string OPTION_FALSE = "no";
	
	public enum OptionState{
		OFF, ON, AUTO;
	}
	
	public interface MultiTypeMap : Gee.Map<string, string>{
		public abstract bool readonly{get; set;}
		
		
		/**
		 * Returns boolean value.
		 * 
		 * @param key option identifier
		 * @return false if value is empty or "off" (case insensitive), true otherwise
		 */
		public bool get_bool(string key, bool @default=false){
			string? val = this[key];
			if(val == null) return @default;
			if(val == OPTION_TRUE) return true;
			if(val == OPTION_FALSE) return false;
			return @default;
		}
		
		public void set_bool(string key, bool val){
			this[key] = val ? OPTION_TRUE : OPTION_FALSE;
		}
		
		/**
		 * Returns integer value.
		 * 
		 * @param key option identifier
		 * @return default value if value is empty or integer value
		 */
		public int get_int(string key, int @default=0,
			int? min=null, int? max=null){
			string? val = this[key];
			if(val == null) return @default;
			int num = int.parse(val);
			if(min != null && num < min) return @default;
			if(max != null && num > max) return @default;
			return num;
		}
		
		public void set_int(string key, int val){
			this[key] = val.to_string();
		}
		
		
		/**
		 * Returns double value.
		 * 
		 * @param key option identifier
		 * @return default value if value is empty or integer value
		 */
		public double get_double(string key, double @default=0.0,
			double? min=null, double? max=null){
			string? val = this[key];
			if(val == null) return @default;
			double num = double.parse(val);
			if(min != null && num < min) return @default;
			if(max != null && num > max) return @default;
			return num;
		}
		
		public void set_double(string key, double val){
			this[key] = val.to_string();
		}
		
		
		public OptionState get_state(string key,
			OptionState @default=OptionState.AUTO){
			int val = this.get_int(key, @default,
				OptionState.OFF, OptionState.AUTO);
			return (OptionState) val;
		}
		
		public void set_state(string key, OptionState val){
			this.set_int(key, (int) val);
		}
		/**
		 * Returns array of strings
		 * @param key option indentifier
		 * @return values
		 */
		public string[] array(string key){
			string[] result = {};
			var line = this[key];
			if(line == null) return result;
			
			string[] parts = line.split(",");
			if(parts.length == 0) return result;
			
			foreach(string s in parts){
				s = s.strip();
				if(s.length > 0) result += (owned) s;
			}
			parts = null;
			return result;
		}
		
		/**
		 * Returns array of strings
		 * @param key option indentifier
		 * @return values
		 */
		public void set_array(string key, string[] array){
			if(array.length == 0){
				this[key] = "";
				return;
			}
			
			this[key] = string.joinv(",", array);
		}
		
		/**
		 * Returns array of strings
		 * @param key option indentifier
		 * @return values
		 */
		public SList<string> get_list(string key){
			var result = new SList<string>();
			var line = this[key];
			if(line == null) return (owned) result;
			
			string[] parts = line.split(",");
			if(parts.length == 0) return result;
			
			foreach(string s in parts){
				s = s.strip();
				
				lib_debug("S: %s", s);
				if(s.length > 0) result.append((owned) s);
			}
			parts = null;
			return result;
		}
		
		
		/**
		 * Returns array of strings
		 * @param key option indentifier
		 * @return values
		 */
		public void set_list(string key, SList<string> list){
			if(list.length() == 0){
				this[key] = "";
				return;
			}
			var builder = new StringBuilder();
			bool first = true;
			foreach(var item in list){
				lib_debug("Config list save: %s", item);
				if(first){
					first = false;
				}
				else{
					builder.append_c(',');
				}
				builder.append(item);
			}
			
			this[key] = builder.str;
		}
		
		/**
		 * Saves configuration to file.
		 * 
		 * @throws GLib.Error on failure
		 */
		
		
		public signal void value_changed(string key);
		
		
		
	}
	
	public interface Serializable: GLib.Object{
		public abstract void save() throws Error;
		public abstract void purge() throws Error;
	}
	
	/**
	 * Class parsing key-value configuration file.
	 * 
	 * Class parser configuration files which contain one key-value pair per line.
	 * Leading and traling white characters are removed.
	 * Key and value are separated by "=" (value may contain "=" character).
	 * Lines starting with a hash "#" are ignored.
	 * {{{
	 * foo = bar
	 * # nice comment
	 * apple = pear
	 * # This is also valid:
	 * comparation = <, =, >
	 * }}}
	 */
	public class MapFile: Gee.HashMap<string, string>, MultiTypeMap, Serializable{
		protected File file;
		public bool readonly{ get; set; }
		/**
		 * Creates new map representation of given file.
		 * 
		 * Creates new map representation of given file. Prints warning on parse errors.
		 * Note: key names are always lowercased.
		 * 
		 * @param file file to parse
		 */
		public MapFile(File file, bool readonly=false){
			this.readonly = readonly;
			this.file = file;
			try{
				var dis = new DataInputStream(file.read());
				string line;
				int lineno = 0;
				while ((line = dis.read_line (null)) != null){
					lineno++;
					string item = line.strip();
					// Ignore empty lines or comments
					if(item.length == 0 || item[0] == '#') continue;
					string[] parts = item.split("=", 2);
					if(parts.length != 2){
						// Implicit keys without value are not supported
						lib_warning("ConfigFile: Parse error at %s:%i",
							file.get_uri(), lineno);
						continue;
					}
					string key = parts[0].strip();
					if(key.length == 0){
						// Empty key
						lib_warning("ConfigFile: Parse error at %s:%i",
							file.get_uri(), lineno);
						continue;
					}
					this[key] = parts[1].strip();
				}
			}
			catch(Error e){
				lib_debug("ConfigFile: Unable to read %s", file.get_uri());
			}
		}
		
		/**
		 * Saves configuration to file.
		 * 
		 * @throws GLib.Error on failure
		 */
		public void save() throws Error{
			if(readonly) throw new IOError.READ_ONLY("Configuration file is opened read-only");
			var builder = new StringBuilder();
			foreach(var e in this.entries){
				builder.append(e.key);
				builder.append(" = ");
				builder.append(e.value);
				builder.append_c('\n');
			}
			string str = builder.str;
			overwrite_file(file, (owned) str);
		}
		
		/**
		 * Purge configuration.
		 * 
		 * @throws GLib.Error on failure
		 */
		public void purge() throws Error{
			this.clear();
			this.save();
		}
		
		/**
		 * Set value.
		 * 
		 * @param key configuration key
		 * @param val value
		 */
		public override void set(string key, string val){
//~ 			lib_debug("%s = %s", key, val);
			if(base.get(key) == val) return;
			base.set(key, val);
			this.value_changed(key);
		}
	}
	
	public class MultiTypeHashMap: Gee.HashMap<string, string>, MultiTypeMap{
		public bool readonly{ get; set; }
	}
}
