/*
 * Copyright 2011-2013 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

namespace Nuvola
{

public class Core
{
	private Diorite.Application app;
	private Diorite.Storage storage;
	private Configuration config;
	private Diorite.Connection connection;
	private FormatSupport formats;
	
	public Core(Diorite.Application app, Diorite.Storage storage, Configuration config, Diorite.Connection connection, FormatSupport formats)
	{
		this.app = app;
		this.storage = storage;
		this.config = config;
		this.connection = connection;
		this.formats = formats;
		var cache_model = Environment.get_variable("NUVOLA_CACHE_MODEL");
		switch (cache_model)
		{
		case "DOCUMENT_VIEWER":
			WebKit.set_cache_model(WebKit.CacheModel.DOCUMENT_VIEWER);
			debug("Cache model: %s", cache_model);
			break;
		case "DOCUMENT_BROWSER":
			WebKit.set_cache_model(WebKit.CacheModel.DOCUMENT_BROWSER);
			debug("Cache model: %s", cache_model);
			break;
		case "WEB_BROWSER":
			WebKit.set_cache_model(WebKit.CacheModel.WEB_BROWSER);
			debug("Cache model: %s", cache_model);
			break;
		default:
			debug("Cache model: default");
			break;
		}
		
		if (config.data_cache)
			connection.enable_cache(config.data_cache_size);
		try
		{
			connection.set_up_proxy(config.proxy_settings,
				config.proxy_type, config.proxy_server,
				config.proxy_port);
		}
		catch (Diorite.ProxyError e)
		{
			warning("Unable to set up proxy: %s", e.message);
		}
		connection.set_up_cookies();
		formats.block_incompatible_flash = config.block_incompatible_flash;
		formats.check_all();
		config.value_changed.connect_after(on_config_changed);
	}
	
	/**
	 * Purge cookies and reload the web interface.
	 */
	public void purge_cookies()
	{ 
		connection.set_up_cookies(true);
	}
	
	/**
	 * Purge settings and show request to restart application.
	 */
	public void purge_settings()
	{
		try
		{
			config.purge();
		}
		catch (Error e)
		{
			warning("Unable to purge configuration: %s", e.message);
		}
		
		var dir = storage.get_child(SERVICES_DIRNAME).user_config_dir;
		try
		{
			Diorite.System.purge_directory_content(dir);
		}
		catch (GLib.Error e)
		{
			warning("Unable to purge services' conf: %s", e.message);
		}
		
		app.show_user_info(
		/// Title in information dialog when default settings were restored
		_("Default settings were restored."),
		/// Request to restart application
		_("Please restart application to finish the process."));
	}
	
	/**
	 * Purge cache and reload the web interface.
	 */
	public async void purge_cache()
	{
		Idle.add(purge_cache.callback);
		yield;
		connection.clear_cache();
		Idle.add(purge_cache.callback);
		yield;
			
		var dir = storage.user_cache_dir;
		try
		{
			Diorite.System.purge_directory_content(dir, true);
		}
		catch(GLib.Error e)
		{
			critical("Unable to purge album art: %s", e.message);
		}
		Idle.add(purge_cache.callback);
		yield;
	}
	
	private void on_config_changed(string key)
	{
		switch(key)
		{
		case Settings.DATA_CACHE:
			if (config.data_cache)
				connection.enable_cache(config.data_cache_size);
			else
				connection.disable_cache();
			break;
		case Settings.DATA_CACHE_SIZE:
			connection.enable_cache(config.data_cache_size);
			break;
		case Settings.PROXY_SETTINGS:
		case Settings.PROXY_TYPE:
		case Settings.PROXY_SERVER:
		case Settings.PROXY_PORT:
			try
			{
				connection.set_up_proxy(config.proxy_settings,
					config.proxy_type, config.proxy_server,
					config.proxy_port);
			}
			catch (Diorite.ProxyError e)
			{
				warning("Unable to set up proxy: %s", e.message);
			}
			break;
		case Settings.BLOCK_INCOMPATIBLE_FLASH:
			formats.block_incompatible_flash = config.block_incompatible_flash;
			formats.check_all();
			break;
		}
	}
}

} // namespace Nuvola
