/*
 * Copyright 2011-2013 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

namespace Nuvola
{

public class FormatSupport: GLib.Object
{
	public static const string MP3_SUPPORT_URL = "http://nuvolaplayer.fenryxo.cz/help/issues/html5_audio.html";
	public static const string FLASH_SUPPORT_URL = "http://nuvolaplayer.fenryxo.cz/help/issues/flash_plugin.html";
	
	public bool flash_available{ get; private set; default = false;}
	public bool mp3_available{ get; private set; default = false;}
	public bool flash_enabled{ get; set; default = true;}
	public bool mp3_enabled{ get; private set; default = true;}
	public bool block_incompatible_flash{ get; set; default = true;}
	private string? mp3_uri;
	private Gst.Plugin? gst_typefind_plugin;
	
	public FormatSupport(string? mp3_uri)
	{
		init_gstreamer();
		this.mp3_uri = mp3_uri;
		gst_typefind_plugin = null;
	}
	
	~FormatSupport()
	{
		if (gst_typefind_plugin != null)
			gst_typefind_plugin.unref();
	}
	public virtual signal void check_all()
	{
		flash_available = check_flash_support();
		mp3_available = check_mp3_support();
	}
	
	/**
	 * Disable GStreamer support
	 * 
	 */
	public void disable_gstreamer()
	{
		#if HAVE_GST_0_10
		var registry = Gst.Registry.get_default();
		#else
		var registry = Gst.Registry.@get();
		#endif
		
		// This is a hack how to make GStreamer non-functional,
		// remove an essential plugin
		var plugin = registry.find_plugin("typefindfunctions");
		if (plugin != null)
		{
			registry.remove_plugin(plugin);
			gst_typefind_plugin = plugin;
			plugin.ref(); // prevents critical warnings
			debug("GStreamer has been disabled.");
		}
		else if(gst_typefind_plugin == null)
		{
			warning("Failed to disable GStreamer");
		}
		
		mp3_enabled = false;
		mp3_available = check_mp3_support();
	}
	
	private void init_gstreamer()
	{
		string[] a = {};
		unowned string[] b = a;
		try
		{
			if (!Gst.init_check(ref b));
				debug("Unable to init %s, maybe already initialized", Gst.version_string());
		}
		catch(Error e)
		{
			warning("Unable to init %s: %s", Gst.version_string(), e.message);
		}
	}
	
	private Gst.Pipeline? create_gst_pipeline(string mp3_uri, out string? error=null)
	{
		error = null;
		try{
			
			#if HAVE_GST_0_10
			var launch_string = "playbin2";
			#else
			// playbin2 became playbin in Gst 1.0
			var launch_string = "playbin";
			#endif
			launch_string += " uri=" + mp3_uri;
			debug("%s launch %s", Gst.version_string(), launch_string);
			var pipeline = Gst.parse_launch(launch_string) as Gst.Pipeline;
			if (pipeline == null)
				error = "Unable to create %s pipeline".printf(Gst.version_string());
			
			return pipeline;
		}
		catch(Error e)
		{
			error = "Failed to create a %s pipeline: %s".printf(Gst.version_string(), e.message);
			return null;
		}
	}
	
	/**
	 * Check whether it's possible to play a MP3 file
	 * 
	 * @param mp3_file		a MP3 file to test
	 * @return true on success, false on failure
	 */
	public bool check_mp3_support()
	{
		debug("Built with Gstreamer %d.%d.%d", GstVersion.MAJOR, GstVersion.MINOR, GstVersion.MICRO);
		Environment.set_variable("GST_DEBUG", "*:2", false);
		string? error_msg;
		if (mp3_uri == null)
			return false;
		var pipeline = create_gst_pipeline(mp3_uri, out error_msg);
		if (pipeline == null)
		{
			warning(error_msg);
			return false;
		}
		
		/* It is not necessary to set Gst.State.PLAYING, because Gstreamer looks for necesary
		 * plugins alsready for Gst.State.PAUSED. */
		switch (pipeline.set_state(Gst.State.PAUSED))
		{
		case Gst.StateChangeReturn.SUCCESS:
			pipeline.set_state(Gst.State.NULL); // free pipeline
			return true;
		case Gst.StateChangeReturn.ASYNC:
			var state_change = pipeline.get_state(null, null, Gst.CLOCK_TIME_NONE);
			pipeline.set_state(Gst.State.NULL); // free pipeline
			if (state_change == Gst.StateChangeReturn.SUCCESS)
				return true;
			
			warning("Unable to change %s pipeline status (async)", Gst.version_string());
			return false;
		default:
			pipeline.set_state(Gst.State.NULL); // free pipeline
			warning("Unable to change %s pipeline status (sync)", Gst.version_string());
			return false;
		}
	}
		
	public bool check_flash_support()
	{
		const string PLUGIN_DIR_KEY = "MOZ_PLUGIN_PATH";
		const string NUVOLA_FLASH_DIR = "/opt/nuvolaplayer/flash/wrapped";
		const string NUVOLA_FLASH_LIBNAME = "/npwrapper.libflashplayer.so";
		const string FLASH_PLUGIN_NAME = "Shockwave Flash";
		
		var plugin_dirs = Environment.get_variable(PLUGIN_DIR_KEY);
		if (plugin_dirs == null || plugin_dirs == ""){
			plugin_dirs = NUVOLA_FLASH_DIR;
		}
		else{
			plugin_dirs = "%s:%s".printf(NUVOLA_FLASH_DIR, plugin_dirs);
		}
		Environment.set_variable(PLUGIN_DIR_KEY, plugin_dirs, true);
		
		
		var flash_available = false;
		var db = WebKit.get_web_plugin_database();
		unowned GLib.SList<WebKit.WebPlugin> plugins = db.get_plugins();
		foreach(var plugin in plugins){
			var path = plugin.get_path();
			var name = plugin.get_name();
			if(block_incompatible_flash){
				if(path.has_suffix(NUVOLA_FLASH_LIBNAME)){
					plugin.enabled = true;
					flash_available = true;
				}
				else{
					plugin.enabled = false;
				}
			}
			else if (FLASH_PLUGIN_NAME in name){
				flash_available = true;
				plugin.enabled = true;
			}
			else{
				plugin.enabled = false;
			}
				
			debug("WebPlugin: %s (%s): %s", name, plugin.enabled.to_string(), path);
		}
		return flash_available;
	}
}
// Gst 1.0 has version constants in a separate header file
#if HAVE_GST_0_10
[CCode(cheader_filename = "gst/gst.h")]
#else
[CCode(cheader_filename = "gst/gstversion.h")]
#endif
namespace GstVersion{
	[CCode(cname="GST_VERSION_MAJOR")]
	public extern const int MAJOR;
	[CCode(cname="GST_VERSION_MINOR")]
	public extern const int MINOR;
	[CCode(cname="GST_VERSION_MICRO")]
	public extern const int MICRO;
}

} // namespace Nuvola
