/*
 * Copyright 2013 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
namespace Nuvola
{

public class UpdatesService: GLib.Object
{
	private Diorite.Application app;
	private Configuration config;
	private ServiceUpdates updates;
	private MainWindow main_window;
	private Diorite.InfoBar infobar = null;
	private uint callback_id = 0;
	
	public UpdatesService(Diorite.Application app, Configuration config, ServiceUpdates updates, MainWindow main_window)
	{
		this.app = app;
		this.config = config;
		this.updates = updates;
		this.main_window = main_window;
		if (config.check_updates)
			start(15);
		config.value_changed.connect(on_settings_changed);
	}
	
	private void start(uint delay_seconds)
	{
		if (callback_id == 0)
			callback_id = Timeout.add_seconds(delay_seconds, check_updates_loop);
	}
	
	private void stop()
	{
		if (callback_id != 0)
		{
			Source.remove(callback_id);
			callback_id = 0;
		}
	}
	
	private bool check_updates_loop()
	{
		callback_id = 0;
		if (config.check_updates)
			updates.check_updates.begin(check_updates_cb);
		
		return false;
	}
	
	private void check_updates_cb(GLib.Object? o, AsyncResult res)
	{
		remove_bar();
		
		int new_services; int new_updates;
		try
		{
			updates.check_updates.end(res, out new_services, out new_updates);
			if (new_services > 0 || new_updates > 0)
			{
				infobar = new Diorite.InfoBar.info("<b>Updates Manager:</b> %d new %s and %d service %s available".printf(
				new_services, new_services == 1 ? "service" : "services",
				new_updates, new_updates == 1 ? "update" : "updates"));
				var update_button = new Gtk.Button.with_label(_("Install"));
				update_button.clicked.connect(on_update_button_clicked);
				infobar.add_button(update_button);
				main_window.box.pack_end(infobar, false);
				infobar.show();
			}
		}
		catch (GLib.Error e)
		{
			warning("Failed to get info about service updates: %s", e.message);
		}
		Timeout.add_seconds(12 * 60 * 60, check_updates_loop);
	}
	
	private void on_update_button_clicked(Gtk.Button button)
	{
		remove_bar();
		show_updates();
	}
	
	private void remove_bar()
	{
		if (infobar != null)
		{
			var parent = infobar.get_parent() as Gtk.Container;
			if (parent != null)
				parent.remove(infobar);
			infobar = null;
		}
	}
	
	public void show_updates()
	{
		var dialog = new ServiceUpdatesDialog(app, updates);
		dialog.run();
		dialog.destroy();
	}
	
	private void on_settings_changed(string key)
	{
		if (key == Settings.CHECK_UPDATES)
		{
			if (config.check_updates)
				start(5);
			else
				stop();
		}
	}
	
}

} // namespace Nuvola
