/*
 * Copyright 2013 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
namespace Nuvola
{

/**
 * Service Updates manage service update packages from remote repository.
 */
public class ServiceUpdates: GLib.Object
{
	private ServicesManager manager;
	private Diorite.Connection conn;
	private string server;
	private string path;
	private HashTable<string, ServiceUpdate>? data;
	/**
	 * Creates new service updates manager
	 * 
	 * @param conn           network connection to use
	 * @param server         root path to the location of updates
	 * @param api_version    JavaScript API
	 */
	public ServiceUpdates(ServicesManager manager, Diorite.Connection conn, string server, string api_version)
	{
		this.manager = manager;
		this.server = server;
		this.path = "%s/api_%s.json".printf(server, api_version);
		this.conn = conn;
		this.data = null;
	}
	
	public async void check_updates(out int new_services, out int new_updates) throws GLib.Error
	{
		var services = manager.list_services();
		new_services = 0;
		new_updates = 0;
		
		yield download_data();
		
		foreach(var key in data.get_keys())
		{
			var service = services.lookup(key);
			if (service == null)
			{
				data[key].is_new = true;
				new_services++;
			}
			else{
				var update = data[key];
				if (update.version_major > service.version
				|| (update.version_major == service.version && update.version_minor > service.version_minor))
				{
					new_updates++;
				}
			}
		}
	}
	
	public void get_updates(out SList<ServiceUpdate> new_services, out SList<ServiceUpdate> updates)
	{
		var services = manager.list_services();
		new_services = null;
		updates = null;
		var keys = data.get_keys();
		keys.sort(strcmp);
		foreach(var key in keys)
		{
			var update = data[key];
			var service = services.lookup(key);
			if (service == null)
			{
				update.is_new = true;
				new_services.append(update);
			}
			else{
				
				if (update.version_major > service.version
				|| (update.version_major == service.version && update.version_minor > service.version_minor))
				{
					update.is_new = false;
					updates.append(update);
				}
			}
		}
	}
	
	public Service install_update(ServiceUpdate update) throws ServiceError
	{
		File tmp_dir;
		try
		{
			tmp_dir = File.new_for_path(DirUtils.make_tmp("nuvolaplayerXXXXXX"));
		}
		catch(FileError e)
		{
			throw new ServiceError.IOERROR(e.message);
		}
		
		try
		{
			var package = tmp_dir.get_child(update.package);
			var uri = "%s/%s".printf(server, update.package);
			if (!conn.download_file(uri, package))
				throw new ServiceError.IOERROR("Cannot download '%s'.", uri);
			
			return manager.install_service(package);
		}
		finally
		{
			Diorite.System.try_purge_dir(tmp_dir);
		}
	}
	
	private async void download_data() throws GLib.Error
	{
		var message = new Soup.Message("GET", path);
		SourceFunc callback = download_data.callback;
		conn.session.queue_message(message, () =>
		{
			Idle.add((owned) callback);
		});
		yield;
		var response = (string) message.response_body.flatten().data;
		var parser = new Json.Parser();
		unowned Json.Node root_node;
		string error_msg;
		if (message.status_code != Soup.KnownStatusCode.OK)
		{
			debug("Server response:\n%s", response);
			try
			{
				parser.load_from_data(response);
			}
			catch (GLib.Error e)
			{
				debug("Parse error: %s", e.message);
				// Probably not a JSON data
			}
			root_node = parser.get_root();
			if (root_node != null && root_node.get_node_type() == Json.NodeType.OBJECT)
			{
				error_msg = js_string_member(root_node.get_object(), "error");
				if (error_msg != null)
					throw new ServiceError.SERVER_ERROR_MESSAGE("Unexpected response from update service: %s", error_msg);
			}
			throw new ServiceError.SERVER_ERROR("Unexpected response from update service: Error code %u - %s.", message.status_code, message.reason_phrase);
		}
		
//~ 		stdout.printf(">>>>>>>>>>>\n%s\n----------\n", response);
		data = new HashTable<string, ServiceUpdate>(str_hash, str_equal);
		
		try
		{
			parser.load_from_data(response);
		}
		catch (GLib.Error e2)
		{
			throw new ServiceError.INVALID_FILE("Unexpected response from update service: Invalid JSON data. %s.", e2.message);
		}
		root_node = parser.get_root();
		if (root_node == null || root_node.get_node_type() != Json.NodeType.OBJECT)
			throw new ServiceError.INVALID_FILE("Unexpected response from update service: Invalid JSON data.");
		
		var services = root_node.get_object();
		error_msg = js_string_member(services, "error");
		if (error_msg != null)
			throw new ServiceError.SERVER_ERROR_MESSAGE("Unexpected response from update service: %s", error_msg);
		
		var members = services.get_members();
		foreach (var key in members)
		{
			var member = services.get_member(key);
			if (member.get_node_type() != Json.NodeType.OBJECT)
			{
				warning("Member '%s' is not an object.", key);
				continue;
			}
			
			var service_data = member.get_object();
			
			var service_id = js_string_member(service_data, "id");
			if (service_id == null)
			{
				warning("Missing 'id' member for '%s'", key);
				continue;
			}
			var name = js_string_member(service_data, "name");
			if (name == null)
			{
				warning("Missing 'name' member for '%s'", key);
				continue;
			}
			var version = js_string_member(service_data, "version");
			if (version == null)
			{
				warning("Missing 'version' member for '%s'", key);
				continue;
			}
			var api = js_string_member(service_data, "api");
			if (api == null)
			{
				warning("Missing 'api' member for '%s'", key);
				continue;
			}
			var changes = js_string_member(service_data, "changes");
			if (changes == null)
			{
				warning("Missing 'changes' member for '%s'", key);
				continue;
			}
			var package = js_string_member(service_data, "package");
			if (package == null)
			{
				warning("Missing 'package' member for '%s'", key);
				continue;
			}
			string[] versions = version.split(".");
			if (versions.length != 2)
			{
				warning("Invalid version member '%s' for '%s'", version, key);
				continue;
			}
			var version_major = int.parse(versions[0]);
			var version_minor = int.parse(versions[1]);
			if (version_major < 1 || version_minor < 0)
			{
				warning("Invalid version member '%s' for '%s'", version, key);
				continue;
			}
			
			data.insert(key, new ServiceUpdate(service_id, name, version_major,
			version_minor, api, changes, package));
		}
	}
}

string? js_string_member(Json.Object object, string key)
{
	if (!object.has_member(key))
		return null;
	var member = object.get_member(key);
	return  (member.get_node_type() == Json.NodeType.VALUE
	&& member.get_value_type() == typeof(string)) ? member.dup_string() : null;
}

public class ServiceUpdate
{
	public string service_id{ get; private set; }
	public string name{ get; private set; }
	public int version_major{ get; private set; }
	public int version_minor{ get; private set; }
	public string api{ get; private set; }
	public string changes{ get; private set; }
	public string package{ get; private set; }
	public bool is_new{ get; set; default = false;}
	
	public ServiceUpdate(string service_id, string name, int version_major,
	int version_minor, string api, string changes, string package)
	{
		this.service_id = service_id;
		this.name = name;
		this.version_major = version_major;
		this.version_minor = version_minor;
		this.api = api;
		this.changes = changes;
		this.package = package;
	}
}

} // namespace Nuvola
