/*
 * Copyright 2011-2013 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#if LASTFM

using Diorite.Logger;

private int64 get_timestamp()
{
	var datetime = new DateTime.now_utc();
	return datetime.to_unix();
}

/**
 * Module providing scrobbling to [[http://www.last.fm/|Last.fm]]
 */
namespace Nuvola.Extensions.Lastfm
{
	
/**
 * Last.fm integration errors
 * 
 * * JSON_PARSE_ERROR - failed to parse Last.fm response
 * * LASTFM_ERROR - invalid Last.fm request
 * * WRONG_RESPONSE - invalid Last.fm response
 */
public errordomain Error
{
	JSON_PARSE_ERROR,
	LASTFM_ERROR,
	WRONG_RESPONSE,
	NO_SESSION
}

const string UI_SPEC_START = """
<ui>
<menubar action="menubar">
<menu action="control">
<placeholder name="control-bottom">
""";

const string UI_SPEC_END = """
</placeholder>
</menu>
</menubar>
</ui>
""";

private unowned Json.Node? json_get_member(Json.Object object, string name)
{
	return object.has_member(name) ? object.get_member(name) : null;
}

private unowned Json.Array? json_get_array(Json.Object object, string name)
{
	unowned Json.Node? node = json_get_member(object, name);
	return node != null && node.get_node_type() == Json.NodeType.ARRAY ? node.get_array() : null;
}

private unowned Json.Object? json_get_object(Json.Object object, string name)
{
	unowned Json.Node? node = json_get_member(object, name);
	return node != null && node.get_node_type() == Json.NodeType.OBJECT ? node.get_object() : null;
}

/**
 * HTTP method used to send request
 */
private enum Method
{
	GET,
	POST;
}

private const string LASTFM_ID = "lastfm";
private const string LASTFM_NAME = "Last.fm";
/**
 * Last.fm API key identifying Nuvola Player. You have to obtain your own
 * for your application, see http://www.last.fm/api/account
 */
private const string LASTFM_API_KEY = "209d19abf80aa2c6d049216ce46830d7";
/**
 * Last.fm secret used to sign requests. You have to obtain your own
 * for your application, see http://www.last.fm/api/account
 */
private const string LASTFM_API_SECRET = "d157730073941bdd851eac950f3154e6";
/**
 * Auth URL of Last.fm API
 */
private const string LASTFM_AUTH_ENDPOINT = "http://www.last.fm/api/auth/";
/**
 * Root URL of Last.fm API
 */
private const string LASTFM_API_ROOT = "http://ws.audioscrobbler.com/2.0/";


private const string LIBREFM_ID = "librefm";
private const string LIBREFM_NAME = "Libre.fm";
private const string LIBREFM_API_KEY = "nuv";
private const string LIBREFM_API_SECRET = "d157730073941bdd851eac950f3154e6";
private const string LIBREFM_API_ROOT = "http://libre.fm/2.0/";
private const string LIBREFM_AUTH_ENDPOINT = "http://libre.fm/api/auth/";
/**
 * Delay before song is scrobbled.
 */
private const int SCROBBLE_SONG_DELAY = 60;
/**
 * Delay before "now playing" status is updated.
 */
private const int NOW_PLAYING_TIMEOUT = 20;

const string SERVICE_ID_KEY = "nuvola_service_id";

const string LOVE_ACTION = "%s-love-toggle";
const string LOVE_KEY = "love";
const string BAN_ACTION = "%s-ban-toggle";
const string BAN_KEY = "ban";

/**
 * Last.fm scrobbling extension
 */
public class Extension: Nuvola.Extension
{
	private weak Diorite.MultiTypeMap config;
	private weak Nuvola.Player player;
	private weak WebBackend web_backend;
	private weak Nuvola.ServicesManager services_manager;
	private weak Diorite.Actions actions;
	private weak Diorite.ActionGroups action_groups;
	private string? current_service_id = null;
	private uint scrobble_track_source = 0;
	private string? song = null;
	private string? artist = null;
	private int64 timestamp = 0;
	private bool can_scrobble = false;
	private Scrobbler[] scrobblers;
	private bool ignore_actions = true;
	private uint ui_merge_id = 0;
	
	construct
	{
		has_preferences = true;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override void load(ObjectContainer objects) throws ExtensionError
	{
		// Get dependencies, throws ExtensionError
		config = objects.get<Diorite.MultiTypeMap>("config");
		player = objects.get<Player>("player");
		web_backend = objects.get<WebBackend>("web_backend"); 
		services_manager = objects.get<ServicesManager>("services_manager");
		actions = objects.get<Diorite.Actions>("actions");
		action_groups = objects.get<Diorite.ActionGroups>("action_groups");
		
		scrobblers = {
			new Scrobbler(LASTFM_ID, LASTFM_NAME, LASTFM_AUTH_ENDPOINT, LASTFM_API_KEY,
			LASTFM_API_SECRET, LASTFM_API_ROOT, true, true, config),
			new Scrobbler(LIBREFM_ID, LIBREFM_NAME, LIBREFM_AUTH_ENDPOINT, LIBREFM_API_KEY,
			LIBREFM_API_SECRET, LIBREFM_API_ROOT, false, false, config)
		};
		
		ignore_actions = true;
		append_ui_actions();
		var service = web_backend.service;
		current_service_id = service == null ? null : service.id;
		on_song_changed(player.song, player.artist, player.album, player.album_art);
		player.song_changed.connect_after(on_song_changed);
		player.notify["playback-state"].connect_after(on_playback_state_changed);
		web_backend.notify["service"].connect_after(on_service_changed);
		foreach (var scrobbler in scrobblers)
			scrobbler.notify["has-session"].connect_after(on_has_session_changed);
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override void unload()
	{
		foreach (var scrobbler in  scrobblers)
		{
			scrobbler.notify["has-session"].disconnect(on_has_session_changed);
			var action_name = LOVE_ACTION.printf(scrobbler.id);
			var action = actions.get_action(action_name);
			if (action != null)
			{
				var toggle_action = action as Diorite.ToggleAction;
				if (toggle_action != null)
					toggle_action.toggled.disconnect(perform_toggle_action);
				action_groups.remove_action_from_group(action, EXTRA_PLAYER_ACTIONS);
				actions.remove_action(action_name);
			}
			
			action_name = BAN_ACTION.printf(scrobbler.id);
			action = actions.get_action(action_name);
			if (action != null)
			{
				var toggle_action = action as Diorite.ToggleAction;
				if (toggle_action != null)
					toggle_action.toggled.disconnect(perform_toggle_action);
				action_groups.remove_action_from_group(action, EXTRA_PLAYER_ACTIONS);
				actions.remove_action(action_name);
			}
		}
		scrobblers = {};
		web_backend.notify["service"].disconnect(on_service_changed);
		player.notify["playback-state"].disconnect(on_playback_state_changed);
		player.song_changed.disconnect(on_song_changed);
		var timestamp = get_timestamp();
		if (can_scrobble && this.song != null && this.artist != null)
		{
			scrobble_track.begin(song, artist, timestamp, (obj, res) =>
			{
				scrobble_track.end(res);
			});
		}
		if (scrobble_track_source != 0)
		{
			Source.remove(scrobble_track_source);
			scrobble_track_source = 0;
		}
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override Gtk.Widget? get_preferences()
	{
		var notebook = new Gtk.Notebook();
		foreach (var scrobbler in  scrobblers)
		{
			var preferences = new ScrobblerPreferences(services_manager, scrobbler);
			preferences.show_all();
			notebook.append_page(preferences, new Gtk.Label(scrobbler.name));
		}
		return notebook;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override void add_actions_ui(Gtk.UIManager manager)
	{
		return_if_fail(actions != null);
		var buffer = new StringBuilder(UI_SPEC_START);
		var have_action = false;
		foreach (var scrobbler in  scrobblers)
		{
			if (scrobbler.can_love)
			{
				buffer.append("<menuitem action=\"");
				buffer.append_printf(LOVE_ACTION, scrobbler.id);
				buffer.append("\" />\n");
				have_action = true;
			}
			if (scrobbler.can_ban)
			{
				buffer.append("<menuitem action=\"");
				buffer.append_printf(BAN_ACTION, scrobbler.id);
				buffer.append("\" />\n");
				have_action = true;
			}
		}
		if (have_action)
		{
			buffer.append(UI_SPEC_END);
			try
			{
				ui_merge_id = manager.add_ui_from_string(buffer.str, buffer.len);
			}
			catch (GLib.Error e)
			{
				warning("[%s] Unable to add ui: %s", id, e.message);
			}
		}
		else
		{
			ui_merge_id = 0;
		}
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override void remove_actions_ui(Gtk.UIManager manager){
		if (ui_merge_id != 0)
		{
			manager.remove_ui(ui_merge_id);
			ui_merge_id = 0;
		}
	}
	
	private void append_ui_actions()
	{
		return_if_fail(actions != null);
		foreach (var scrobbler in  scrobblers)
		{
			Diorite.ToggleAction action;
			if (scrobbler.can_love)
			{
				action = new Diorite.ToggleAction(LOVE_ACTION.printf(scrobbler.id),
				null, _("%s: Love song").printf(scrobbler.name), null, null, true);
				action.toggled.connect(perform_toggle_action);
				actions.add_action(action);
				action_groups.add_action_to_group(action, EXTRA_PLAYER_ACTIONS);
				scrobbler.set_data<Diorite.ToggleAction>(LOVE_KEY, action);
			}
			
			if (scrobbler.can_ban)
			{
				action = new Diorite.ToggleAction(BAN_ACTION.printf(scrobbler.id),
				null, _("%s: Ban song").printf(scrobbler.name), null, null, true);
				action.toggled.connect(perform_toggle_action);
				actions.add_action(action);
				action_groups.add_action_to_group(action, EXTRA_PLAYER_ACTIONS);
				scrobbler.set_data<Diorite.ToggleAction>(BAN_KEY, action);
			}
		}
	}
	
	private void perform_toggle_action(Gtk.ToggleAction toggle_action)
	{
		if (ignore_actions)
			return;
		
		// Cannot use foreach loop, because it produces
		// invalid C code with valac 0.14.2.
		for (var i = 0; i < scrobblers.length; i++)
		{
			Scrobbler scrobbler = scrobblers[i];
			
			var action = scrobbler.get_data<Diorite.ToggleAction>(LOVE_KEY);
			if (action != null && action == toggle_action)
			{
				scrobbler.set_loved_flag.begin(player.song, player.artist, toggle_action.active, (obj, res) =>
				{
					try
					{
						scrobbler.set_loved_flag.end(res);
					}
					catch (Error e)
					{
						warning("Scrobbler (%s) error: %s", scrobbler.name, e.message);
					}
				});
				break;
			}
			action = scrobbler.get_data<Diorite.ToggleAction>(BAN_KEY);
			if (action != null && action == toggle_action)
			{
				scrobbler.set_banned_flag.begin(player.song, player.artist, toggle_action.active, (obj, res) =>
				{
					try
					{
						scrobbler.set_loved_flag.end(res);
					}
					catch (Error e)
					{
						warning("Scrobbler (%s) error: %s", scrobbler.name, e.message);
					}
				});
				break;
			}
		}
	}
	
	/**
	 * Callback for song change. Queues "now playing" update and track scrobbling.
	 */
	private void on_song_changed(string? song, string? artist, string? album, string? album_art)
	{
		var timestamp = get_timestamp();
		
		if (scrobble_track_source != 0)
		{
			Source.remove(scrobble_track_source);
			scrobble_track_source = 0;
		}
		
		if (song == null || artist == null)
			return;
		
		if (can_scrobble && this.song != null && this.artist != null)
		{
			scrobble_track.begin(this.song, this.artist, this.timestamp, (obj, res) =>
			{
				scrobble_track.end(res);
			});
		}
		
		this.song = song;
		this.artist = artist;
		this.timestamp = timestamp;
		can_scrobble = false;
		
		update_now_playing();
		
		scrobble_track_source = Timeout.add_seconds(SCROBBLE_SONG_DELAY, enable_scrobbling_cb);
		
		update_actions.begin(song, artist, (obj, res) =>
		{
			update_actions.end(res);
		});
	}
	
	private void on_playback_state_changed(GLib.Object o, ParamSpec p)
	{
		update_now_playing();
	}
	
	private bool enable_scrobbling_cb()
	{
		if (scrobble_track_source != 0)
		{
			can_scrobble = true;
			scrobble_track_source = 0;
		}
		
		return false;
	}
	
	private async void scrobble_track(string song, string artist, int64 timestamp)
	{
		if (current_service_id == null)
			return;
		foreach (var scrobbler in scrobblers)
		{
			if (scrobbler.is_service_enabled(current_service_id))
				{
				try
				{
					yield scrobbler.scrobble_track(song, artist, timestamp);
				}
				catch (Error e)
				{
					warning("Scrobbler (%s) error: %s", scrobbler.name, e.message);
				}
			}
		}
	}
	
	private async void update_actions(string? song, string? artist)
	{
		var invalidate = current_service_id == null || song == null || artist == null;
		foreach (var scrobbler in scrobblers)
		{
			var action = scrobbler.get_data<Diorite.ToggleAction>(LOVE_KEY);
			if (action != null)
			{
				ignore_actions = true;
				if (invalidate || !scrobbler.has_session || !scrobbler.can_love)
				{
					action.sensitive = false;
					action.active = false;
				}
				else
				{
					try
					{
						action.active = yield scrobbler.is_loved(song, artist);
					}
					catch (Error e)
					{
						warning("Scrobbler (%s) error: %s", scrobbler.name, e.message);
						action.active = false;
					}
					action.sensitive = true;
				}
				ignore_actions = false;
			}
			
			action = scrobbler.get_data<Diorite.ToggleAction>(BAN_KEY);
			if (action != null)
			{
				ignore_actions = true;
				if (invalidate || !scrobbler.has_session || !scrobbler.can_ban)
				{
					action.sensitive = false;
					action.active = false;
				}
				else
				{
					try
					{
						action.active = yield scrobbler.is_banned(song, artist);
					}
					// Cannot use variable e, because it produces invalid C code
					// with valac 0.14.2
					catch (Error e2)
					{
						warning("Scrobbler (%s) error: %s", scrobbler.name, e2.message);
						action.active = false;
					}
					action.sensitive = true;
				}
				ignore_actions = false;
			}
		}
	}
	
	private async void update_now_playing()
	{
		if (current_service_id == null || player.playback_state != STATE_PLAYING
		|| player.artist == null || player.song == null)
			return;
		
		foreach (var scrobbler in scrobblers)
		{
			if (scrobbler.is_service_enabled(current_service_id))
			{
				try
				{
					yield scrobbler.update_now_playing(player.song, player.artist);
				}
				catch (Error e)
				{
					warning("Scrobbler (%s) error: %s", scrobbler.name, e.message);
				}
			}
		}
	}
	
	private void on_service_changed(GLib.Object o, ParamSpec p)
	{
		var service = web_backend.service;
		current_service_id = service == null ? null : service.id;
	}
	
	private void on_has_session_changed(GLib.Object o, ParamSpec p)
	{
		update_actions(player.song, player.artist);
	}
}

public class Scrobbler: GLib.Object
{
	public string id { get; private set; }
	public string name { get; private set; }
	public bool can_love { get; private set; }
	public bool can_ban { get; private set; }
	public bool has_session { get; private set; default = false; }
	private string api_key;
	private string api_secret;
	private string api_root;
	private string auth_endpoint;
	private string? session;
	private string? token;
	private string username_key;
	private string session_key;
	private string services_key;
	private string? username = null;
	private weak Diorite.MultiTypeMap config;
	private SList<string> banned_tracks = null;
	private bool banned_tracks_set = false;
	
	public Scrobbler(string id, string name, string auth_endpoint, string api_key,
	string api_secret, string api_root, bool can_love, bool can_ban, Diorite.MultiTypeMap config)
	{
		this.id = id;
		this.name = name;
		this.auth_endpoint = auth_endpoint;
		this.api_key = api_key;
		this.api_secret = api_secret;
		this.api_root = api_root;
		this.can_love = can_love;
		this.can_ban = can_ban;
		this.config = config;
		username_key =  "%s.%s.username".printf(EXTENSIONS_KEY, id);
		session_key =  "%s.%s.session_key".printf(EXTENSIONS_KEY, id);
		services_key =  "%s.%s.services.".printf(EXTENSIONS_KEY, id);
		session = config[session_key];
		if (session == "")
			session = null;
		if (session != null)
		{
			has_session = true;
			username = config[username_key];
			if (username == "")
				username = null;
		}
	}
	
	/**
	 * Whether client has request token to obtain session key
	 * @return true if client has request token
	 */
	public bool has_token()
	{
		return token != null;
	}
	
	/**
	 * Forgets request token
	 */
	public void drop_token()
	{
		token = null;
	}
	
	/**
	 * Generates authorization URL to authorize request token
	 * 
	 * @return authorization URL
	 * @throws Error on failure
	 */
	public async string get_auth_url() throws Error
	{
		token = yield get_request_token();
		return generate_auth_url(token);
	}
	
	/**
	 * Generates URL of authorization page
	 * 
	 * This is the second step of desktop application authentication. User has to
	 * authorize provided request token. 
	 * See http://www.last.fm/api/desktopauth for details.
	 * 
	 * @param token request token obtained by get_request_token()
	 * @return url of the authorization page
	 */
	public string generate_auth_url(string token)
	{
		return "%s?api_key=%s&token=%s".printf(auth_endpoint, api_key, token);
	}
	
	/**
	 * Exchanges authorized request token for session key.
	 * 
	 * @throws Error on failure
	 */
	public async void fetch_session() throws Error
	{
		session = yield get_session_key(token);
		token = null;
		config[session_key] = session ?? "";
		config[username_key] = username ?? "";
		save_config();
		has_session = session != null;
	}
	
	/**
	 * Forgets session key
	 */
	public void drop_session()
	{
		session = null;
		has_session = false;
		config[session_key] = "";
		save_config();
	}
	
	private void save_config()
	{
		var serializable = config as Diorite.Serializable;
		if (serializable != null)
		{
			try
			{
				serializable.save();
			}
			catch (GLib.Error e)
			{
				warning("Unable to save configuration: %s", e.message);
			}
		}
	}
	
	public bool is_service_enabled(string service_id)
	{
		return config.get_bool(services_key + service_id, true);
	}
	
	public void set_service_enabled(string service_id, bool enabled)
	{
		config.set_bool(services_key + service_id, enabled);
	}
	
	/**
	 * Scrobbles track to Last.fm
	 * 
	 * @param song song name
	 * @param artist artist name
	 * @param timestamp Unix time
	 * @throws Error on failure
	 */
	public async void scrobble_track(string song, string artist, int64 timestamp) throws Error
	{
		if (session == null)
			return;
		debug("%s scrobble: %s by %s, %s", id, song, artist, timestamp.to_string());
		// http://www.last.fm/api/show/track.scrobble
		var params = new HashTable<string,string>(null, null);
		params.insert("method", "track.scrobble");
		params.insert("api_key", api_key);
		params.insert("sk", session);
		params.insert("track", song);
		params.insert("artist", artist);
		params.insert("timestamp", timestamp.to_string());
	
		var response = yield send_request(Method.POST, params);
		if (!response.has_member("scrobbles"))
			throw new Error.WRONG_RESPONSE("Response doesn't contain scrobbles member.");
	}
	
	/**
	 * Updates now playing status on Last.fm
	 * 
	 * @param song song name
	 * @param artist artist name
	 * @throws Error on failure
	 */
	public async void update_now_playing(string song, string artist) throws Error
	{
		if (session == null)
			return;
		const string API_METHOD = "track.updateNowPlaying";
		debug("%s update now playing: %s by %s", id, song, artist);
		// http://www.last.fm/api/show/track.updateNowPlaying
		var params = new HashTable<string,string>(null, null);
		params.insert("method", API_METHOD);
		params.insert("api_key", api_key);
		params.insert("sk", session);
		params.insert("track", song);
		params.insert("artist", artist);
	
		var response = yield send_request(Method.POST, params);
		if (!response.has_member("nowplaying"))
			throw new Error.WRONG_RESPONSE("%s: Response doesn't contain nowplaying member.", API_METHOD);
	}
	
	public async bool is_loved(string song, string artist) throws Error
	{
		if (session == null)
			return false;
		if (username == null)
			yield retrieve_username();
		const string API_METHOD = "track.getInfo";
		debug("%s love/ban status: %s by %s", id, song, artist);
		// http://www.last.fm/api/show/track.getInfo
		var params = new HashTable<string,string>(null, null);
		params.insert("method", API_METHOD);
		params.insert("api_key", api_key);
		params.insert("sk", session);
		params.insert("track", song);
		params.insert("artist", artist);
		params.insert("username", username);
		params.insert("autocorrect", "1");
	
		var response = yield send_request(Method.POST, params);
		if (!response.has_member("track"))
			throw new Error.WRONG_RESPONSE("%s: Response doesn't contain track member.", API_METHOD);
		
		var track = response.get_object_member("track");
		if (!track.has_member("userloved"))
			throw new Error.WRONG_RESPONSE("%s: Response doesn't contain userloved member.", API_METHOD);
		var userloved = track.get_string_member("userloved");
		return userloved != "0";
	}
	
	public async void set_loved_flag(string song, string artist, bool loved) throws Error
	{
		// http://www.last.fm/api/show/track.love
		// http://www.last.fm/api/show/track.unlove
		yield set_flag(song, artist, loved ? "track.love" : "track.unlove");
	}
	
	public async bool is_banned(string song, string artist) throws Error
	{
		if (!banned_tracks_set)
		{
			yield get_banned_tracks();
			banned_tracks_set = true;
		}
		var track = song + "::" + artist;
		return banned_tracks.find_custom(track, strcmp) != null;
	}
	
	private async void get_banned_tracks() throws Error
	{
		const string API_METHOD = "user.getBannedTracks";
		if (session == null)
			throw new Error.NO_SESSION("%s: There is no authorized session.", API_METHOD);
		if (username == null)
			yield retrieve_username();
		
		banned_tracks = null;
		int pages = 1;
		for (int page = 1; page <= pages; page++)
		{
			// http://www.last.fm/api/show/user.getBannedTracks
			var params = new HashTable<string,string>(null, null);
			params.insert("method", API_METHOD);
			params.insert("api_key", api_key);
			params.insert("sk", session);
			params.insert("user", username);
			params.insert("page", page.to_string());
			params.insert("limit", "20");
		
			var response = yield send_request(Method.GET, params);
			if (!response.has_member("bannedtracks"))
				throw new Error.WRONG_RESPONSE("%s: Response doesn't contain bannedtracks member.", API_METHOD);
			var bannedtracks = response.get_object_member("bannedtracks");
			if (bannedtracks.has_member("totalPages") && bannedtracks.get_string_member("totalPages") == "0")
				break; // no banned tracks
			if (!bannedtracks.has_member("@attr"))
				throw new Error.WRONG_RESPONSE("%s: Response doesn't contain @attr member.", API_METHOD);
			var attr = bannedtracks.get_object_member("@attr");
			if (!attr.has_member("totalPages"))
				throw new Error.WRONG_RESPONSE("%s: Response doesn't contain totalPages member.", API_METHOD);
			var pages_str = attr.get_string_member("totalPages");
			pages = int.parse(pages_str);
			if (pages <= 0)
			{
				warning("%s: Invalid number of pages '%s'.", id, pages_str);
				pages = -1; // do not continue next cycle
			}
			if (!bannedtracks.has_member("track"))
				throw new Error.WRONG_RESPONSE("%s: Response doesn't contain track member.", API_METHOD);
			
			unowned Json.Object? banned_track = json_get_object(bannedtracks, "track");
			if (banned_track != null)
			{
				add_banned_track(banned_track);
			}
			else
			{
				var tracks = json_get_array(bannedtracks, "track");
				if (tracks != null)
				{
					var size =  tracks.get_length();
					for (uint i = 0; i < size; i++)
						add_banned_track(tracks.get_object_element(i));
				}
			}
		}
	}
	
	private void add_banned_track(Json.Object track)
	{
		if (!track.has_member("name") || !track.has_member("artist"))
		{
			warning("%s: Banned track doesn't contain name or artist member.'", id);
			return;
		}
		var artist = track.get_object_member("artist");
		if (!artist.has_member("name"))
		{
			warning("%s: Banned track's artist doesn't contain name member.'", id);
			return;
		}
		var song = track.get_string_member("name") + "::" + artist.get_string_member("name");
		message("banned song %s", song);
		banned_tracks.prepend(song);
	}
	
	public async void set_banned_flag(string song, string artist, bool banned) throws Error
	{
		// http://www.last.fm/api/show/track.ban
		// http://www.last.fm/api/show/track.unban
		yield set_flag(song, artist, banned ? "track.ban" : "track.unban");
		var track = song + "::" + artist;
		if (banned)
		{
			banned_tracks.prepend(track);
		}
		else
		{
			unowned SList<string> node;
			while ((node = banned_tracks.find_custom(track, strcmp)) != null)
				banned_tracks.delete_link(node);
		}
	}
	
	private async void set_flag(string song, string artist, string flag) throws Error
	{
		if (session == null)
			throw new Error.NO_SESSION("%s: There is no authorized session.", flag);
		var params = new HashTable<string,string>(null, null);
		params.insert("method", flag);
		params.insert("api_key", api_key);
		params.insert("sk", session);
		params.insert("track", song);
		params.insert("artist", artist);
	
		var response = yield send_request(Method.POST, params);
		if (!response.has_member("status"))
			throw new Error.WRONG_RESPONSE("%s: Response doesn't contain status member.", flag);
		if (response.get_string_member("status") != "ok")
			throw new Error.WRONG_RESPONSE("%s: Response doesn't contain ok status.", flag);
	}
	
	private async void retrieve_username() throws Error
	{
		const string API_METHOD = "user.getInfo";
		if (session == null)
			throw new Error.NO_SESSION("%s: There is no authorized session.", API_METHOD);
		debug("%s get user info", id);
		// http://www.last.fm/api/show/user.getInfo
		var params = new HashTable<string,string>(null, null);
		params.insert("method", API_METHOD);
		params.insert("api_key", api_key);
		params.insert("sk", session);
		var response = yield send_request(Method.GET, params);
		if (!response.has_member("user"))
			throw new Error.WRONG_RESPONSE("%s: Response doesn't contain user member.", API_METHOD);
		var user = response.get_object_member("user");
		if (!user.has_member("name"))
			throw new Error.WRONG_RESPONSE("%s: Response doesn't contain name member.", API_METHOD);
		username = user.get_string_member("name");
		if (username == null || username == "")
			throw new Error.WRONG_RESPONSE("%s: Response contains empty username.", API_METHOD);
	}
	
	
	/**
	 * Obtains request token
	 * 
	 * This is the first step of desktop application authentication.
	 * See http://www.last.fm/api/desktopauth for details.
	 * 
	 * @return request token
	 * @throws Error on failure
	 */
	private async string get_request_token() throws Error
	{
		// http://www.last.fm/api/show/auth.getToken
		const string API_METHOD = "auth.getToken";
		var params = new HashTable<string,string>(null, null);
		params.insert("method", API_METHOD);
		params.insert("api_key", api_key);
		
		var response = yield send_request(Method.GET, params);
		if (!response.has_member("token"))
			throw new Error.WRONG_RESPONSE("%s: Response doesn't contain token member.", API_METHOD);
		
		return response.get_string_member("token");
	}
	
	/**
	 * Obtains session key
	 * 
	 * This is the third step of desktop application authentication. Session key is used
	 * to authorize Last.fm API requests.
	 * See http://www.last.fm/api/desktopauth for details.
	 * 
	 * @param request_token request token obtained by get_request_token()
	 * @return session token
	 * @throws Error on failure
	 */
	private async string get_session_key(string request_token) throws Error
	{
		// http://www.last.fm/api/show/auth.getSession
		const string API_METHOD = "auth.getSession";
		var params = new HashTable<string,string>(null, null);
		params.insert("method", API_METHOD);
		params.insert("api_key", api_key);
		params.insert("token", request_token);
		
		var response = yield send_request(Method.GET, params);
		if (!response.has_member("session"))
			throw new Error.WRONG_RESPONSE("%s: Response doesn't contain session member.", API_METHOD);
		
		var session = response.get_object_member("session");
		if (!session.has_member("key"))
			throw new Error.WRONG_RESPONSE("%s: Response doesn't contain session.key member.", API_METHOD);
		
		if (session.has_member("name"))
			username = session.get_string_member("name");
		return session.get_string_member("key");
	}
	
	/**
	 * Send Last.fm API request
	 * 
	 * @param method HTTP method to use to send request
	 * @param params Last.fm API parameters of request
	 * @return Root JSON object of the response
	 * @throws Error on failure
	 */
	private async Json.Object send_request(Method method, HashTable<string,string> params) throws Error
	{
		Soup.Message message;
		var request = create_signed_request(params) + "&format=json";
		if (method == Method.GET)
		{
			message = new Soup.Message("GET", api_root + "?" + request);
		}
		else if (method == Method.POST)
		{
			message = new Soup.Message("POST", api_root);
			message.set_request("application/x-www-form-urlencoded",
				Soup.MemoryUse.COPY, request.data);
		}
		else
		{
			message = null;
			error("Last.fm: Unsupported request method: %d", method);
		}
		
		var session = WebKit.get_default_session();
		SourceFunc callback = send_request.callback;
		session.queue_message(message, () =>
		{
			Idle.add((owned) callback);
		});
		yield;
		
		var response = (string) message.response_body.flatten().data;
//~ 		stdout.printf(">>>>>>>>>>>\n%s\n---------\n%s\n----------\n", (string) request.data, response);
		var parser = new Json.Parser();
		try
		{
			parser.load_from_data(response);
		}
		catch (GLib.Error e)
		{
			throw new Error.JSON_PARSE_ERROR(e.message);
		}
		
		var root_object = parser.get_root().get_object();
		if (root_object.has_member("error") && root_object.has_member("message"))
		{
			throw new Error.LASTFM_ERROR("%s: %s".printf(
				root_object.get_int_member("error").to_string(),
				root_object.get_string_member("message")
				));
		}
		return root_object;
	}
	
	/**
	 * Creates signed request string for Last.fm API call
	 * 
	 * @param params parameters of the request
	 */
	private string create_signed_request(HashTable<string,string> params)
	{
		// See http://www.last.fm/api/desktopauth#6
		
		// Buffer for request string
		var req_buffer = new StringBuilder();
		// Buffer to compute signature for request
		var sig_buffer = new StringBuilder();
		
		// Signature requires sorted params
		var keys = params.get_keys();
		keys.sort(strcmp);
				
		foreach(var key in keys){
			var val = params.lookup(key);
			// signature buffer does not contain "=" and "&"
			// to separate key and value or key-value pairs
			// TODO: how about escaping?
			sig_buffer.append(key);
			sig_buffer.append(val);
			
			// request buffer contains "=" and "&"
			// to separate key and value or key-value pairs
			append_param(req_buffer, key, val);
		}
		
		// Append API_SECRET and generate MD5 hash
		sig_buffer.append(api_secret);
		var api_sig = Checksum.compute_for_string(ChecksumType.MD5, sig_buffer.str);
		sig_buffer.truncate();
		
		// Append signature to the request string
		append_param(req_buffer, "api_sig", api_sig);
		return req_buffer.str;
	}
	
	/**
	 * Appends URL parameter
	 * 
	 * Appends URL parameter in format "key=value" or "&key=value".
	 * 
	 * @param buffer Buffer which parameter will be appended to
	 * @param key parameter name
	 * @param value parameter value
	 */
	public void append_param(StringBuilder buffer, string key, string value)
	{
		if (buffer.len > 0)
		{
			buffer.append_c('&');
		}
		buffer.append(Uri.escape_string(key, "", true));
		buffer.append_c('=');
		buffer.append(Uri.escape_string(value, "", true)); 
	}
}

public class ScrobblerPreferences: Gtk.Grid
{
	private Scrobbler scrobbler;
	private weak ServicesManager services_manager;
	
	public ScrobblerPreferences(ServicesManager services_manager, Scrobbler scrobbler)
	{
		Object(orientation: Gtk.Orientation.VERTICAL, row_spacing: 10);
		this.services_manager = services_manager;
		this.scrobbler = scrobbler;
		margin = 10;
		update();
	}
	
	private void update()
	{
		foreach (var child in get_children())
			remove(child);
		
		Gtk.Label label;
		Gtk.Button button;
		
		if (scrobbler.has_session)
		{
			label = new Gtk.Label(_("Scrobbling to %s is enabled.").printf(scrobbler.name));
			add(label);
			button = new Gtk.Button.with_label(_("Unauthorize"));
			button.clicked.connect(() => {
				scrobbler.drop_session();
				update();
			});
			add(button);
			
			var grid = new Gtk.Grid();
			grid.column_spacing = 10;
			grid.row_spacing = 1;
			var services = services_manager.list_services();
			var keys = services.get_keys();
			keys.sort(strcmp);
			var line_number = 0;
			foreach (var key in keys)
			{
				var service = services.lookup(key);
				if (service == null)
					continue;
				var service_name = new Gtk.Label(service.name);
				service_name.vexpand = true;
				grid.attach(service_name, 0, line_number, 1, 1);
				var toggle = new Gtk.Switch();
				toggle.active = scrobbler.is_service_enabled(service.id);
				toggle.set_data<string?>(SERVICE_ID_KEY, service.id);
				toggle.notify["active"].connect_after(on_service_toggled);
				grid.attach(toggle, 1, line_number, 1, 1);
				line_number++;
			}
			add(grid);
		}
		else if(scrobbler.has_token())
		{
			label = new Gtk.Label(_("%s authorization page should have been opened in your web browser. Allow acces to Nuvola Player and click on the button bellow to finish authorization process.").printf(scrobbler.name));
			add(label);
			button = new Gtk.Button.with_label(_("Complete authorization"));
			button.clicked.connect(() =>
			{
				button.sensitive = false;
				scrobbler.fetch_session.begin((obj, res) =>
				{
					try{
						scrobbler.fetch_session.end(res);
						update();
					}
					catch (Lastfm.Error e)
					{
						warning("Scrobbler (%s) error: %s", scrobbler.name, e.message);
						button.label = _("Error occurred. Try again.");
					}
					button.sensitive = true;
				});
			});
			add(button);
		}
		else
		{
			label = new Diorite.WrappedLabel(_("You have to grant access to Nuvola Player to enable scrobbling to %s. Authorization page will be opened in your web browser.").printf(scrobbler.name));
			add(label);
			button = new Gtk.Button.with_label(_("Authorize"));
			button.clicked.connect(() =>
			{
				button.sensitive = false;
				scrobbler.get_auth_url.begin((obj, res) =>
				{
					try
					{
						var url = scrobbler.get_auth_url.end(res);
						Diorite.System.open_uri_with_fallback(url);
						update();
					}
					catch (Lastfm.Error e)
					{
						warning("Scrobbler (%s) error: %s", scrobbler.name, e.message);
						button.label = _("Error occurred. Try again.");
					}
					button.sensitive = true;
				});
			});
			add(button);
		}
		
		label.xpad = 0;
		label.use_markup = true;
		label.wrap_mode = Pango.WrapMode.WORD;
		label.wrap = true;
		label.justify = Gtk.Justification.FILL;
			
		show_all();
	}
	
	private void on_service_toggled(GLib.Object o, ParamSpec p)
	{
		var toggle = o as Gtk.Switch;
		var service_id = o.get_data<string?>(SERVICE_ID_KEY);
		return_if_fail(toggle != null);
		scrobbler.set_service_enabled(service_id, toggle.active);
	}
}

public Nuvola.ExtensionInfo get_info(){
	return {
		/// Name of a plugin providing scrobbling to Last.fm
		_("Last.fm and Libre.fm"),
		Config.VERSION,
		/// Extension descriptiom
		_("<p>This plugin sends information about currently playing and recently played song (<i>scrobbling</i>) to the <a href=\"http://www.last.fm\">Last.fm</a> and <a href=\"http://www.libre.fm\">Libre.fm</a> services. They can then examine your music taste to recommend you similar songs.</p>"),
		"Jiří Janoušek",
		typeof(Extension),
		true
	};
}

} // namespace Nuvola.Extensions.Lastfm

#endif
