/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Diorite.Logger;
using Diorite;
using Nuvola.Settings;

namespace Nuvola.Extensions.MediaKeys{
const string CONF_ID = "media_keys";
const string MULTIMEDIA_KEYS = EXTENSIONS_KEY + "." + CONF_ID + "." + "multimedia_keys";
const string PAUSE_KEY = EXTENSIONS_KEY + "." + CONF_ID + "." + "pause_key";

public Nuvola.ExtensionInfo get_info(){
	return {
		/// Name of a plugin providing integration with multimedia keys in GNOME
		_("Media keys"),
		Config.VERSION,
		/// Description of a plugin providing integration with multimedia keys in GNOME
		_("<p>This extension allows one to control playback with multimedia keys.</p>"),
		"Jiří Janoušek",
		typeof(Extension),
		true
	};
}

public const string MEDIA_PLAY = "XF86AudioPlay";
public const string MEDIA_PAUSE = "XF86AudioPause";
public const string MEDIA_STOP = "XF86AudioStop";
public const string MEDIA_PREV = "XF86AudioPrev";
public const string MEDIA_NEXT = "XF86AudioNext";
public const string KEY_PAUSE = "Pause";
/**
 * Manages multimedia keys and pause key
 */
public class Extension: Nuvola.Extension{
	private KeybindingManager keybindings;
	private GnomeMedia media_keys;
	private weak Diorite.Application app;
	private weak Player player;
	private weak MultiTypeMap config;	
	
	public bool multimedia_keys{
		get{
			return config.get_bool(MULTIMEDIA_KEYS, true);
		}
		set{
			config.set_bool(MULTIMEDIA_KEYS, value);
		}
	}
	
	public bool pause_key{
		get{
			return config.get_bool(PAUSE_KEY, false);
		}
		set{
			config.set_bool(PAUSE_KEY, value);
		}
	}
	
	construct{
		has_preferences = true;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override void load(ObjectContainer objects) throws ExtensionError{
		// Get dependencies, throws ExtensionError
		app = objects.get<Diorite.Application>("application");
		player = objects.get<Player>("player");
		config = objects.get<Diorite.MultiTypeMap>("config");
		config.value_changed.connect(config_changed);
		config_changed(PAUSE_KEY);
		config_changed(MULTIMEDIA_KEYS);
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override void unload(){
		config.value_changed.disconnect(config_changed);
		release_pause_key();
		release_multimedia_keys();
		keybindings = null;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override Gtk.Widget? get_preferences(){
		var box = new Gtk.Grid();
		box.orientation = Gtk.Orientation.VERTICAL;
		/// Settings dialog - tab General
		var media_keys = new Gtk.CheckButton.with_label(_("Handle multimedia keys."));
		media_keys.active = this.multimedia_keys;
		media_keys.toggled.connect(() => { this.multimedia_keys = media_keys.active; });
		box.add(media_keys);
		
		/// Settings dialog - tab General
		var pause_key = new Gtk.CheckButton.with_label(_("Use Pause (break) key to toggle play/pause"));
		pause_key.active = this.pause_key;
		pause_key.toggled.connect(() => { this.pause_key = pause_key.active; });
		box.add(pause_key);
		box.show_all();
		return box;
	}
	
	private void config_changed(string key){
		switch(key){
		case PAUSE_KEY:
			if(pause_key){
				handle_pause_key();
			}
			else{
				release_pause_key();
			}
			break;
		case MULTIMEDIA_KEYS:
			if(multimedia_keys){
				handle_multimedia_keys();
			}
			else{
				release_multimedia_keys();
			}
			break;
		}
	}
	/**
	 * Binds Pause key to play/pause action
	 */
	private void handle_pause_key(){
		if(this.keybindings == null) this.keybindings = new KeybindingManager();
		this.keybindings.bind(KEY_PAUSE, () => {this.player.toggle_play();});
	}

	/**
	 * Unbinds Pause key from play/pause action
	 */
	private void release_pause_key(){
		if(this.keybindings != null) this.keybindings.unbind(KEY_PAUSE);
	}
	
	/**
	 * Binds multimedia keys to control playback
	 */
	private void handle_multimedia_keys(){

		/* Watch sound menu */
		Bus.watch_name(BusType.SESSION, "org.gnome.SettingsDaemon",
			BusNameWatcherFlags.NONE,
			this.gnome_settings_appeared, this.gnome_settings_vanished);

	}
		
	/**
	 * Unbinds multimedia keys to stop controlling playback
	 */
	private void release_multimedia_keys(){

		this.media_keys_stop_fallback();
		if(this.media_keys == null) return;
		try{
			this.media_keys.release_media_player_keys(app.app_name);
			this.media_keys.media_player_key_pressed.disconnect(this.on_media_key_pressed);
			this.media_keys = null;
			
		}
		catch(IOError e){
			warning("Unable to get proxy for GNOME Media keys: %s", e.message);
			this.media_keys = null;
		}

	}
	
	/**
	 * Use GNOME settings daemon to control multimedia keys
	 */
	private void gnome_settings_appeared(DBusConnection conn, string name, string owner){
		debug("GNOME settings daemon appeared: %s, %s", name, owner);
		media_keys_stop_fallback();
		if(!try_gnome_keys()){
			media_keys = null;
			media_keys_fallback();
		}
	}
	
	private bool try_gnome_keys(){
		try{
			this.media_keys = Bus.get_proxy_sync(BusType.SESSION,
			"org.gnome.SettingsDaemon",
			"/org/gnome/SettingsDaemon/MediaKeys");
			/* Vala includes "return false" if DBus method call fails! */
			this.media_keys.grab_media_player_keys(app.app_name, 0);
			this.media_keys.media_player_key_pressed.connect(this.on_media_key_pressed);
			return true;
			
		}
		catch(IOError e){
			warning("Unable to get proxy for GNOME Media keys: %s", e.message);
			return false;
		}
	}
	
	private void gnome_settings_vanished(DBusConnection conn, string name){
		debug("GNOME settings daemon vanished: %s", name);
		this.media_keys = null;
		this.media_keys_fallback();
	}
	
	/**
	 * Triggers media keys actions
	 */
	private void on_media_key_pressed(string app_name, string key){
		debug("Media key pressed: %s, %s", app_name, key);
		if(app_name != app.app_name) return;
		switch(key){
			case "Play":
			case "Pause": this.player.toggle_play(); break;
			case "Stop": this.player.stop(); break;
			case "Previous": this.player.previous_song(); break;
			case "Next": this.player.next_song(); break;
		}
	}

	/**
	 * Fallback to use Xorg keybindings
	 */
	private void media_keys_fallback(){
		if(this.keybindings == null) this.keybindings = new KeybindingManager();
		this.keybindings.bind(MEDIA_PLAY, () => {this.player.toggle_play();});
		this.keybindings.bind(MEDIA_PAUSE, () => {this.player.toggle_play();});
		this.keybindings.bind(MEDIA_STOP, () => {this.player.stop();});
		this.keybindings.bind(MEDIA_PREV, () => {this.player.previous_song();});
		this.keybindings.bind(MEDIA_NEXT, () => {this.player.next_song();});
	}
	
	private void media_keys_stop_fallback(){
		if(this.keybindings == null) return;
		this.keybindings.unbind(MEDIA_PLAY);
		this.keybindings.unbind(MEDIA_PAUSE);
		this.keybindings.unbind(MEDIA_STOP);
		this.keybindings.unbind(MEDIA_PREV);
		this.keybindings.unbind(MEDIA_NEXT);
	}
}


[DBus(name = "org.gnome.SettingsDaemon.MediaKeys")]
public interface GnomeMedia: Object{
	public abstract void grab_media_player_keys(string app, uint32 time) throws IOError;
	public abstract void release_media_player_keys(string app) throws IOError;
	public signal void media_player_key_pressed(string app, string key);
}

} // namespace Nuvola.Extensions.MediaKeys
