/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Diorite.Logger;

namespace Nuvola.Extensions.TrayIcon{

public Nuvola.ExtensionInfo get_info(){
	return {
		/// Name of a plugin providing scrobbling to Last.fm
		_("Tray Icon"),
		Config.VERSION,
		/// Extension descriptiom
		_("<p>This plugin shows tray icon with menu. Tray icon may be required for <i>hide on close</i> feature.</p>"),
		"Jiří Janoušek",
		typeof(Extension),
		true
	};
}

/**
 * Tray icon is used to control playback when the main window is not visible
 * (it's minimized, hidden or covered by other windows) and to bring the hidden main
 * window to foreground.
 * 
 * The tray icon is shown only if Ubuntu Sound menu is not detected.
 */
public class Extension: Nuvola.Extension{
	private weak Diorite.Application app;
	private weak Diorite.Actions actions;
	private weak Diorite.ActionGroups action_groups;
	private weak Nuvola.MainWindow main_window;
	private weak Nuvola.Player player;
	private Gtk.Menu? menu;
	private Gtk.StatusIcon? icon;
	private Gtk.Widget[] extra_actions = {};
	private uint update_id = 0;
	private int player_actions_count = 0;
	
	/**
	 * {@inheritDoc}
	 */
	public override void load(ObjectContainer objects) throws ExtensionError{
		// Get dependencies, throws ExtensionError
		app = objects.get<Diorite.Application>("application");
		main_window = objects.get<Nuvola.MainWindow>("main_window");
		actions = objects.get<Diorite.Actions>("actions");
		action_groups = objects.get<Diorite.ActionGroups>("action_groups");
		player = objects.get<Nuvola.Player>("player");
		
		icon = new Gtk.StatusIcon.from_icon_name(app.icon);
		icon.title = app.display_name;
		update_tooltip();
		icon.popup_menu.connect(show_menu);
		icon.activate.connect(() => {
			app.activate();
		});
		
		menu = new Gtk.Menu();
		var group = action_groups.get_actions_of_group(BASE_PLAYER_ACTIONS);
		player_actions_count = group.length;
		foreach (var action in group)
			menu.append(create_menu_item(action));
		
		var quit_action = actions.get_action("quit");
		if (quit_action != null)
			menu.append(create_menu_item(quit_action));
		else
			critical("Action '%s' not found.", "quit");
		
		foreach (var action in action_groups.get_actions_of_group(EXTRA_PLAYER_ACTIONS))
			action.notify["sensitive"].connect(on_actions_changed);
		
		action_groups.add_action_to_group.connect_after(on_action_added);
		action_groups.remove_action_from_group.connect(on_action_removed);
		
		update_menu();
		player.song_changed.connect(on_song_changed);
	}
	
	private void on_action_added(Diorite.Action action, string group)
	{
		if (group != EXTRA_PLAYER_ACTIONS)
			return;
		action.notify["sensitive"].connect(on_actions_changed);
		schedule_update();
	}
	
	private void on_action_removed(Diorite.Action action, string group)
	{
		if (group != EXTRA_PLAYER_ACTIONS)
			return;
		action.notify["sensitive"].disconnect(on_actions_changed);
		schedule_update();
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override void unload()
	{
		foreach(var action in action_groups.get_actions_of_group(EXTRA_PLAYER_ACTIONS))
			action.notify["sensitive"].disconnect(on_actions_changed);

		player.song_changed.disconnect(on_song_changed);
		icon.visible = false;
		icon = null;
		menu = null;
	}
	
	private void on_actions_changed(Object o, ParamSpec p)
	{
		schedule_update();
	}
	
	private void schedule_update()
	{
		if(update_id != 0)
			Source.remove(update_id);
		update_id = Idle.add(update_menu_cb);
	}
	
	private bool update_menu_cb()
	{
		update_id = 0;
		update_menu();
		return false;
	}
	
	private void update_menu()
	{
		return_if_fail(menu != null);
		foreach (var item in extra_actions)
			menu.remove(item);
		extra_actions = {};
		int index = player_actions_count;
		foreach (var action in action_groups.get_actions_of_group(EXTRA_PLAYER_ACTIONS))
		{
			if (action.sensitive)
			{
				var menu_item = create_menu_item(action);
				menu.insert(menu_item, index++);
				extra_actions += menu_item;
			}
		}
	}
	
	private Gtk.MenuItem create_menu_item(Diorite.Action action)
	{
		Gtk.MenuItem item;
		var toggle_action = action as Diorite.ToggleAction;
		if (toggle_action != null)
		{
			var check_item = new Gtk.CheckMenuItem();
			check_item.active = toggle_action.active;
			item = check_item;
			toggle_action.notify["active"].connect_after((o, p) =>
			{
				var active = toggle_action.active;
				if (check_item.active != active)
					schedule_update();
			});
		}
		else if(action.stock_id != null)
		{
			var image_item = new Gtk.ImageMenuItem();
			image_item.image = new Gtk.Image.from_icon_name(action.stock_id, Gtk.IconSize.MENU);
			action.notify["stock-id"].connect((o, p) =>
			{
				if (action.stock_id != null)
				{
					image_item.image = new Gtk.Image.from_icon_name(action.stock_id, Gtk.IconSize.MENU);
				}
			});
			item = image_item;
		}
		else
		{
			item = new Gtk.MenuItem();
		}
		
		item.label = action.label.replace("_", "");
		action.notify["label"].connect((o, p) =>
		{
			item.label = action.label.replace("_", "");
		});
		
		item.sensitive = action.sensitive;
		action.notify["sensitive"].connect((o, p) =>
		{
			item.sensitive = action.sensitive;
		});
		item.activate.connect(() => { action.activate(); });
		return item;
	}
	
	private void show_menu(uint button, uint time){
		return_if_fail(menu != null);
		menu.show_all();
		menu.popup(null, null, icon.position_menu, button, time);
	}
	
	private void on_song_changed(string? song, string? artist, string? album, string? album_art){
		update_tooltip();
	}
	
	private void update_tooltip(){
		if(player.playback_state == STATE_NONE || player.song == null){
			icon.tooltip_text = app.display_name;
		}
		else if(player.artist == null){
			icon.tooltip_text = "%s - %s".printf(player.song, app.display_name);
		}
		else{
			icon.tooltip_text = "%s - %s - %s".printf(player.song, player.artist, app.display_name);
		}
	}
}

} // namespace Nuvola.Extensions.TrayIcon
