/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Diorite;
using Diorite.System;

namespace Nuvola{

public enum PreferencesTab{
	GENERAL,
	NETWORK,
	SERVICE,
	EXTENSIONS,
	ADVANCED
}

public class PreferencesWindow : Gtk.Dialog{
	/// Preferences dialog title
	private const string TITLE = _("Preferences");
	private weak Diorite.Application app;
	private weak Nuvola.Configuration config;
	private weak Diorite.Actions actions;
	private weak ExtensionsManager extensions;
	private Gtk.RadioButton proxy_buttons[3];
	private Gtk.RadioButton hide_window_buttons[3];
	private Gtk.Entry proxy_server;
	private Gtk.Entry proxy_port;
	private Gtk.Entry data_cache_size;
	private Gtk.RadioButton proxy_type[2];
	private ObjectContainer objects;
	
	/**
	 * Constructs new main window
	 * 
	 * @param app Application object
	 */
	public PreferencesWindow(ObjectContainer objects, WebBackend web_backend, Diorite.Application app, Storage storage, Connection connection, ServicesManager manager, ExtensionsManager extensions, PreferencesTab default_tab=PreferencesTab.GENERAL){
		this.app = app;
		this.objects = objects;
		
		try{
			config = objects.get<Nuvola.Configuration>("config");
			actions = objects.get<Diorite.Actions>("actions");
		}
		catch(ExtensionError e){
			error("Unable to find configuration object: %s", e.message);
		}
		this.extensions = extensions;
		window_position = Gtk.WindowPosition.CENTER;
		this.title = TITLE;
		this.border_width = 5;
		try {
			icon = Gtk.IconTheme.get_default().load_icon(app.icon, 48, 0);
		}
		catch(Error e){
			warning("Unable to load application icon.");
		}
		
		set_default_size(400, -1);
		
		Gtk.Label label;
		var notebook = new Gtk.Notebook();
		notebook.border_width = 5; 
		var content_box = get_content_area() as Gtk.Box;
		
		var box = new Gtk.Grid();
		box.orientation = Gtk.Orientation.VERTICAL;
		box.border_width = 5; 
		box.show();
		/// Settings dialog, tab General
		notebook.append_page(box, new Gtk.Label(_("General")));
		
		/// Settings dialog - tab General			
		label = new Gtk.Label(_("Run in background when the main window is closed"));
		label.xalign = 0.0f;
		label.margin_top = 10;
		box.add(label);
		this.hide_window_buttons = new Gtk.RadioButton[3];
		/// Settings dialog - tab General
		this.hide_window_buttons[0] = new Gtk.RadioButton.with_label(null, _("Always"));
		box.add(this.hide_window_buttons[0]);
		/// Settings dialog - tab General
		this.hide_window_buttons[1] = new Gtk.RadioButton.with_label(this.hide_window_buttons[0].get_group(), _("Only if music is playing"));
		box.add(this.hide_window_buttons[1]);
		/// Settings dialog - tab General
		this.hide_window_buttons[2] = new Gtk.RadioButton.with_label(this.hide_window_buttons[0].get_group(), _("Never"));
		box.add(this.hide_window_buttons[2]);
		
		switch(config.hide_window){
			case OptionState.ON: this.hide_window_buttons[0].active = true; break;
			case OptionState.OFF: this.hide_window_buttons[2].active = true; break;
			default: this.hide_window_buttons[1].active = true; break;
		}			
		
		this.hide_window_buttons[0].toggled.connect(this.hide_window_toggled);
		this.hide_window_buttons[1].toggled.connect(this.hide_window_toggled);
		this.hide_window_buttons[2].toggled.connect(this.hide_window_toggled);
		
		/// Settings dialog - tab General
		var space_key = new Gtk.CheckButton.with_label(_("Enable keybindings with Space key."));
		space_key.active = config.space_key;
		space_key.toggled.connect(() => {
			config.space_key = space_key.active;
		});
		space_key.margin_top = 10;
		box.add(space_key);
		
		/// Settings dialog - tab General
		var dark_theme = new Gtk.CheckButton.with_label(_("Prefer dark theme."));
		dark_theme.active = config.dark_theme;
		dark_theme.toggled.connect(() => { config.dark_theme = dark_theme.active; });
		box.add(dark_theme);
		
		box = new Gtk.Grid();
		box.orientation = Gtk.Orientation.VERTICAL;
		box.border_width = 10; 
		box.show();
		/// Settings dialog tab
		notebook.append_page(box, new Gtk.Label(_("Network")));
		
		/// Settings dialog, tab Advanced
		var box2 = new Gtk.Grid();
		box2.orientation = Gtk.Orientation.VERTICAL;
		box2.border_width = 10;
		var label2 = new Gtk.Label(null);
		label2.xalign = 0;
		label2.wrap_mode = Pango.WrapMode.WORD;
		label2.wrap = true;
		label2.justify = Gtk.Justification.FILL;
		box2.add(label2);
		
		var service = web_backend.service;
		if (service != null)
		{
			var service_settings = web_backend.get_settings();
			if (service_settings != null)
			{
				/// Label in service-specific settings, %s will be replaced by service name
				label2.set_text(_("These preferences are specific for service %s.").printf(service.name));
				label2.margin_bottom = 10;
				box2.add(service_settings);
			}
			else
			{
				/// Label in service-specific settings
				label2.set_text(_("Service %s does not have any extra preferences.").printf(service.name));
			}
		}
		else
		{
			/// Label in service-specific settings
			label2.set_text(_("No service is currently loaded."));
		}
		
		
		box2.show_all();
		/// Settings dialog tab
		notebook.append_page(box2, new Gtk.Label(_("Service")));
		
		/// Settings dialog - tab Network
		label = new Gtk.Label(_("Some cloud music services don't work with automatic settings and may require manual SOCKS proxy settings. New proxy settings is applied after restart of Nuvola Player."));
		label.xalign = 0;
		label.use_markup = true;
		label.wrap_mode = Pango.WrapMode.WORD;
		label.wrap = true;
		label.justify = Gtk.Justification.FILL;
		box.add(label);
					
		if(!Connection.have_socks()){
			/// Settings dialog - tab Network, shown if tsocks library was not detected
			label = new Gtk.Label(_("<b>Note:</b> <a href='%s'>TSOCKS library</a> was not detected. SOCKS proxy may not be supported.").printf("http://tsocks.sourceforge.net/"));
			label.xalign = 0;
			label.use_markup = true;
			label.wrap_mode = Pango.WrapMode.WORD;
			label.wrap = true;
			label.justify = Gtk.Justification.FILL;
			box.add(label);
		}
					
		this.proxy_buttons = new Gtk.RadioButton[3];
		/// Settings dialog - tab Network
		this.proxy_buttons[0] = new Gtk.RadioButton.with_label(null, _("Use direct connection without a proxy server"));
		box.add(this.proxy_buttons[0]);
		proxy_buttons[0].margin_top = 10;
		
		/// Settings dialog - tab Network
		this.proxy_buttons[1] = new Gtk.RadioButton.with_label(this.proxy_buttons[0].get_group(), _("Use system HTTP(S) proxy settings"));
		box.add(this.proxy_buttons[1]);
		
		/// Settings dialog - tab Network
		this.proxy_buttons[2] = new Gtk.RadioButton.with_label(this.proxy_buttons[0].get_group(), _("Use manual proxy settings"));
		box.add(this.proxy_buttons[2]);
		
		var table = new Gtk.Grid();
		table.column_spacing = 5;
		box.add(table);
		int line = 0;
		
		
		table.attach(new Gtk.Label(_("Proxy Type")), 0, line, 1, 1);
		//line + 1, opt, opt, 5, 5);
		this.proxy_type = new Gtk.RadioButton[2];
		/// Settings dialog - tab Network, proxy type HTTP(S)
		this.proxy_type[0] = new Gtk.RadioButton.with_label(null, _("HTTP(S)"));
		table.attach(this.proxy_type[0], 1, line, 1, 1);
		//, opt, opt, 5, 5);
		/// Settings dialog - tab Network, proxy type SOCKS
		this.proxy_type[1] = new Gtk.RadioButton.with_label(this.proxy_type[0].get_group(), _("SOCKS"));
		table.attach(this.proxy_type[1], 2, line, 1, 1);
		//, opt, opt, 5, 5);
		if(config.proxy_type == ProxyType.SOCKS)
			this.proxy_type[1].active = true;
		else
			this.proxy_type[0].active = true;
		this.proxy_type[0].toggled.connect(this.proxy_type_toggled);
		this.proxy_type[1].toggled.connect(this.proxy_type_toggled);
		line++;
		/// Settings dialog - tab Network
		table.attach(new Gtk.Label(_("Proxy Server")), 0, line, 1, 1);
		// opt, opt, 5, 5);
		this.proxy_server = new Gtk.Entry();
		this.proxy_server.text = config.proxy_server ?? "";
		this.proxy_server.changed.connect(() => {
			config.proxy_server = this.proxy_server.text;
		});
		table.attach(this.proxy_server, 1, line, 2, 1);
		//, opt, opt, 5, 5);
		line++;
		/// Settings dialog - tab Network
		table.attach(new Gtk.Label(_("Proxy Server Port")), 0, line, 1, 1);
		//, opt, opt, 5, 5);
		this.proxy_port = new Gtk.Entry();
		this.proxy_port.text = config.proxy_port ?? "";
		this.proxy_port.changed.connect(() => { 
			config.proxy_port = this.proxy_port.text;
		});
		table.attach(this.proxy_port, 1, line, 2, 1);
		//, opt, opt, 5, 5);
		line++;
		
		var use_proxy = config.proxy_settings;
		if(use_proxy == ProxySettings.MANUAL){
			this.proxy_buttons[2].active = true;
			this.proxy_server.sensitive = true;
			this.proxy_port.sensitive = true;
			this.proxy_type[0].sensitive = true;
			this.proxy_type[1].sensitive = true;
		}
		else if(use_proxy == ProxySettings.AUTO){
			this.proxy_buttons[1].active = true;
			this.proxy_server.sensitive = false;
			this.proxy_port.sensitive = false;
			this.proxy_type[0].sensitive = false;
			this.proxy_type[1].sensitive = false;
		}
		else{
			this.proxy_buttons[0].active = true;
			this.proxy_server.sensitive = false;
			this.proxy_port.sensitive = false;
			this.proxy_type[0].sensitive = false;
			this.proxy_type[1].sensitive = false;
		}
		
		this.proxy_buttons[0].toggled.connect(this.proxy_toggled);
		this.proxy_buttons[1].toggled.connect(this.proxy_toggled);
		this.proxy_buttons[2].toggled.connect(this.proxy_toggled);
		
		
		var extensions_view = new ExtensionsManagerView(extensions, config);
		extensions_view.show();
		/// Label of a tab in preferences window
		notebook.append_page(extensions_view, new Gtk.Label(_("Extensions")));
		
		box = new Gtk.Grid();
		box.orientation = Gtk.Orientation.VERTICAL;
		box.border_width = 10;
		box.show();
		/// Settings dialog tab
		notebook.append_page(box, new Gtk.Label(_("Advanced")));
		
		var gst_key = "gstreamer_" + web_backend.get_libs().lookup("gstreamer_major_minor");
		/// Settings dialog, tab Advanced
		var gstreamer = new Gtk.CheckButton.with_label(_("Load GStreamer to provide HTML5 Audio support"));
		gstreamer.active = config.get_bool(gst_key, true);
		gstreamer.toggled.connect(() => { 
			config.set_bool(gst_key, gstreamer.active);
			if (gstreamer.active)
				web_backend.enable_gstreamer();
			else
				web_backend.disable_gstreamer();
		});
		box.add(gstreamer);
		
		/// Settings dialog, tab Advanced
		var check_updates = new Gtk.CheckButton.with_label(_("Automatically check service updates"));
		check_updates.active = config.check_updates;
		check_updates.toggled.connect(() => { 
			config.check_updates = check_updates.active;
		});
		box.add(check_updates);
		
		/// Settings dialog, tab Advanced
		var web_plugins = new Gtk.CheckButton.with_label(_("Block unnecessary web plugins"));
		web_plugins.active = config.web_plugins_blocker;
		web_plugins.toggled.connect(() => { 
			config.web_plugins_blocker = web_plugins.active;
		});
		box.add(web_plugins);
		
		/// Settings dialog, tab Advanced
		var block_incompatible_flash = new Gtk.CheckButton.with_label(_("Block incompatible Flash plugins"));
		block_incompatible_flash.active = config.block_incompatible_flash;
		block_incompatible_flash.toggled.connect(() => { 
			config.block_incompatible_flash = block_incompatible_flash.active;
		});
		box.add(block_incompatible_flash);
		
		/// Settings dialog, tab Advanced
		var data_cache = new Gtk.CheckButton.with_label(_("Use cache for remote data"));
		data_cache.active = config.data_cache;
		data_cache.toggled.connect(() => { 
			config.data_cache = data_cache.active;
			data_cache_size.sensitive = data_cache.active;
		});
		box.add(data_cache);
		
		var hbox = new Gtk.Grid();
		hbox.column_spacing = 5;
		hbox.orientation = Gtk.Orientation.HORIZONTAL;
		/// Settings dialog, tab Advanced
		hbox.add(new Gtk.Label(_("Maximal data cache size (MB):")));
		this.data_cache_size = new Gtk.Entry();
		this.data_cache_size.text = config.data_cache_size.to_string();
		data_cache_size.sensitive = config.data_cache;
		this.data_cache_size.changed.connect(() => {
			config.data_cache_size = int.parse(data_cache_size.text);
		});
		data_cache_size.hexpand = true;
		hbox.add(data_cache_size);
		box.add(hbox);
		
		string[] ACTIONS = {Actions.PURGE_CACHE, Actions.PURGE_UI_STATE, Actions.PURGE_COOKIES, Actions.PURGE_SETTINGS};
		foreach(var action_id in ACTIONS){
			var action = actions.get_action(action_id);
			if(action != null){
				box.add(new Diorite.ActionButton(action));
			}
		}
		
		content_box.pack_start(notebook, false, true, 0);
		content_box.show_all();
		add_buttons(Gtk.Stock.CLOSE, Gtk.ResponseType.CLOSE, Gtk.Stock.HELP, Gtk.ResponseType.HELP);
		
		notebook.page = (int)default_tab;
	}
	
	#if DEBUG_MEMORY
	~PreferencesWindow(){
		debug("SettingsWindow destroyed");
	}
	#endif
	
	private void proxy_toggled(Gtk.ToggleButton button){
		if(button.active){
			if(button == this.proxy_buttons[0]){
				config.proxy_settings = ProxySettings.NONE;
				this.proxy_server.sensitive = false;
				this.proxy_port.sensitive = false;
				this.proxy_type[0].sensitive = false;
				this.proxy_type[1].sensitive = false;
			}
			else if(button == this.proxy_buttons[1]){
				config.proxy_settings = ProxySettings.AUTO;
				this.proxy_server.sensitive = false;
				this.proxy_port.sensitive = false;
				this.proxy_type[0].sensitive = false;
				this.proxy_type[1].sensitive = false;
			}
			else if(button == this.proxy_buttons[2]){
				config.proxy_settings = ProxySettings.MANUAL;
				this.proxy_server.sensitive = true;
				this.proxy_port.sensitive = true;
				this.proxy_type[0].sensitive = true;
				this.proxy_type[1].sensitive = true;
			}
		}
	}
	
	private void hide_window_toggled(Gtk.ToggleButton button){
		if(button.active){
			if(button == this.hide_window_buttons[0]){
				config.hide_window = OptionState.ON;
			}
			else if(button == this.hide_window_buttons[1]){
				config.hide_window = OptionState.AUTO;
			}
			else if(button == this.hide_window_buttons[2]){
				config.hide_window = OptionState.OFF;
			}
		}
	}
	
	private void proxy_type_toggled(Gtk.ToggleButton button){
		if(button.active){
			if(button == this.proxy_type[0]){
				config.proxy_type = ProxyType.HTTP;
			}
			else if(button == this.proxy_type[1]){
				config.proxy_type = ProxyType.SOCKS;
			}
		}
	}
	
	/**
	 * Close Preferences dialog and save configuration.
	 */
	public new void close(){
		try{
			debug("Close Preferences dialog, save configuration");
			config.save();
		}
		catch(Error e){
			critical("Unable to save configuration: %s", e.message);
		}
		destroy();
	}
	
	public override bool delete_event(Gdk.EventAny event){
		return false;
	}
}

} // namespace Nuvola
