/*
 * Copyright 2011-2013 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

namespace Nuvola{

public class ServicesManagerView: Gtk.Table{
	/**
	 * Name of file with service description in HTML.
	 */
	private static const string DESCRIPTION_FILENAME = "description.html";
	private static const int ICON_SIZE = 26; 
	private Gtk.TreeView selector;
	private ServicesManager manager;
	private HashTable<string, Service> services;
	private Diorite.SimpleDocView page;
	public Gtk.ButtonBox buttons{ get; private set;}
	private File template;
	private string default_icon_name;
	private Gtk.Button button_remove;
	private Gtk.Button use_service_button;
	private Gtk.Button cancel_button;
	public Service? service {get; private set;}
	public string cancel_button_text
	{
		get
		{
			return cancel_button.label;
		}
		set
		{
			cancel_button.label = value;
		}
	}
	
	public signal void response(int response_id);
	
	public ServicesManagerView(string default_icon_name, File template, ServicesManager manager, string? selected_id=null){
		this.manager = manager;
		this.services = manager.list_services();
		this.template = template;
		this.default_icon_name = default_icon_name;
		
		/// Label of cancel button
		cancel_button = new Gtk.Button.with_label(_("Cancel"));
		/// Button on service selector screen
		use_service_button = new Gtk.Button.with_label(_("Use this service"));
		
		// id, pixbuf, name
		var model = new Gtk.ListStore(SMI.COUNT, typeof(string), typeof(Gdk.Pixbuf), typeof(string));
		
		var opts = Gtk.AttachOptions.SHRINK;
		
		var win = new Gtk.ScrolledWindow(null, null);
		selector = new Gtk.TreeView();
		selector.headers_visible = false;
		var column = new Gtk.TreeViewColumn();
		var icon_renderer = new Gtk.CellRendererPixbuf();
		icon_renderer.xpad = 5;
		var name_renderer = new Gtk.CellRendererText();
		name_renderer.ypad = 10;
		column.pack_start(icon_renderer, false);
		column.set_attributes(icon_renderer, "pixbuf", SMI.ICON); 
		column.pack_start(name_renderer, true);
		column.set_attributes(name_renderer, "markup", SMI.NAME);
		column.sort_column_id = SMI.NAME;
		selector.append_column(column);
		
		win.add(selector);
		win.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC);
		win.shadow_type = Gtk.ShadowType.IN;
		border_width = 3;
		win.border_width = 3;
		attach(win, 0, 1, 0, 1, Gtk.AttachOptions.FILL, Gtk.AttachOptions.FILL|Gtk.AttachOptions.EXPAND, 0, 0);
		
		var manager_buttons = new Gtk.HButtonBox();
		manager_buttons.layout_style = Gtk.ButtonBoxStyle.SPREAD;
		attach(manager_buttons, 0, 1, 1, 2, Gtk.AttachOptions.FILL, opts, 0, 0);
		
		var button_add = new Gtk.Button.from_stock(Gtk.Stock.ADD);
		button_add.sensitive = manager.allow_management;
		button_add.clicked.connect(on_add_service_clicked);
		manager_buttons.add(button_add);
		
		button_remove = new Gtk.Button.from_stock(Gtk.Stock.REMOVE);
		button_remove.sensitive = false;
		button_remove.clicked.connect(on_remove_service_clicked);
		manager_buttons.add(button_remove);
		
		page = new Diorite.SimpleDocView();
		page.wrap_mode = Gtk.WrapMode.WORD_CHAR;
		page.border_width = 15;
		
		var page_scroll = new Gtk.ScrolledWindow(null, null);
		page_scroll.shadow_type = Gtk.ShadowType.IN;
		page_scroll.add(page);
		
		attach(page_scroll, 1, 2, 0, 1,
			Gtk.AttachOptions.EXPAND|Gtk.AttachOptions.FILL,
			Gtk.AttachOptions.EXPAND|Gtk.AttachOptions.FILL, 3, 3);
		
		buttons = new Gtk.HButtonBox();
		buttons.layout_style = Gtk.ButtonBoxStyle.SPREAD;
		attach(buttons, 1, 2, 1, 2, Gtk.AttachOptions.FILL|Gtk.AttachOptions.EXPAND, Gtk.AttachOptions.SHRINK, 0, 0);
		
		add_button(cancel_button, Gtk.ResponseType.CANCEL);
		add_button(use_service_button, Gtk.ResponseType.OK);
		
		var selection = selector.get_selection();
		selection.changed.connect(on_selection_changed);
		selection.set_mode(Gtk.SelectionMode.SINGLE);
		selector.row_activated.connect(on_row_activated);
		
		load(model, selected_id);
	}
	
	#if DEBUG_MEMORY
	~ServicesManagerView(){
		debug("~ServicesManagerView()");
	}
	#endif
	
	private void load(Gtk.ListStore model, string? selected_id=null){
		string? selected = null;
		if(selected_id != null){
			selected = selected_id;
		}
		else if(this.service != null){
			selected = this.service.id;
		}
		
		var file = template.get_parent().get_child("_resources").get_child("nuvolaplayer.png");
		Gdk.Pixbuf? default_icon = null;
		try{
			default_icon = Gtk.IconTheme.get_default().load_icon(default_icon_name, ICON_SIZE, 0);
		}
		catch(Error e){
			warning("Unable to load application icon.");
		}
		
		var keys = services.get_keys();
		keys.sort(strcmp);
		Gtk.TreeIter iter;
		Gtk.TreeIter? current_service = null;
		foreach(var key in keys){
			var service = services.lookup(key);
			return_if_fail(service != null);
			model.append(out iter);
			file = service.resources.get_child("icon.svg");
			Gdk.Pixbuf pixbuf;
			try{
				pixbuf = new Gdk.Pixbuf.from_file_at_size(file.get_path(), ICON_SIZE, ICON_SIZE);
			}
			catch(GLib.Error e){
				try{
					file = service.resources.get_child("icon.png");
					pixbuf = new Gdk.Pixbuf.from_file_at_size(file.get_path(), ICON_SIZE, ICON_SIZE);
				}
				catch(GLib.Error e){
					pixbuf = default_icon;
				}
			}
			model.set(iter, SMI.ID, key, SMI.ICON, pixbuf, SMI.NAME,
				"<span size=\"16000\"> %s</span>".printf(Markup.escape_text(service.name)));
			if(service.id == selected){
				current_service = iter;
			}
		}
		
		selector.model = model;
		
		var selection = selector.get_selection();
		if(current_service != null || model.get_iter_first(out current_service)){
			selection.select_iter(current_service);
		}
		selector.grab_focus();
	}
	
	public void reload(string? selected_id=null){
		var model = selector.model as Gtk.ListStore;
		selector.model = null;
		model.clear();
		services = manager.list_services();
		load(model, selected_id);
	}
	
	public void on_row_activated(Gtk.TreePath path, Gtk.TreeViewColumn column){
		var model = selector.model as Gtk.ListStore;
		Gtk.TreeIter iter;
		string id;
		if(model.get_iter(out iter, path)){
			model.get(iter, SMI.ID, out id);
			service = services.lookup(id);
			/* Idle callback prevents gdk_window_get_cursor: assertion `GDK_IS_WINDOW (window)' failed */
			Idle.add(response_ok_cb);
		}
	}
	
	private bool response_ok_cb(){
		if (service != null)
			response(Gtk.ResponseType.OK);
		return false;
	}
	
	public void add_button(Gtk.Button button, int response_id = Gtk.ResponseType.NONE){
		if(response_id != Gtk.ResponseType.NONE){
			button.clicked.connect(() => {
				response(response_id);
			});
		}
		buttons.add(button);
	}
	
	private void on_selection_changed(Gtk.TreeSelection selection)
	{
		var model = selector.model as Gtk.ListStore;
		Gtk.TreeIter iter;
		string id;
		if (selection.get_selected(null, out iter))
		{
			model.get(iter, SMI.ID, out id);
			service = services.lookup(id);
			if (service != null)
			{
				button_remove.sensitive = service.removable;
				load_service_page(service);
			}
			else
			{
				button_remove.sensitive = false;
			}
			use_service_button.sensitive = service != null;
		}
	}
	
	private void load_service_page(Service service){
		var builder = new StringBuilder();
		var version = "%d.%d".printf(service.version, service.version_minor);
		var api_version = "%d.%d".printf(service.api_major, service.api_minor);
		var maintainer_name = Markup.escape_text(service.maintainer_name);
		var maintainer_link = Markup.escape_text(service.maintainer_link);
		var resources = service.resources.get_uri() + "/";
		var path = service.resources.get_path();
		
		string desc = null;
		var description_file = service.resources.get_child(DESCRIPTION_FILENAME);
		if (description_file.query_file_type(0) == FileType.REGULAR){
			try{
				desc = Diorite.System.read_file(description_file);
			}
			catch(Error e){
				warning("Unable to read description: %s", e.message);
			}
		}
		
		if (desc == null)
			desc = _("No description provided for this service.");
		
		string requirements;
		string link_text;
		int faq;
		if (service.requirements_specified)
		{
			requirements = "";
		}
		else
		{
			if(service.flash_plugin){
				/// Warning in service selector if service requires Flash
				requirements = _("This streaming service requires an Adobe Flash plugin compatible with GTK+ 3 to work properly.");
				link_text = "How to install compatible Flash plugin";
				faq = 2107;
			}
			else{
				/// Message in service selector if service doesn't require Flash
				requirements = _("This streaming service requires a MP3 decoder for Gstreamer to work properly.");
				faq = 2277;
				link_text = "How to install HTML5 Audio support";
			}
			
			requirements = "<h2>Playback Requirements</h2>\n<p>%s <a href=\"%s%d\">%s</a>.</p>".printf(
			Markup.escape_text(requirements), "https://answers.launchpad.net/nuvola-player/+faq/",
			faq, Markup.escape_text(link_text));
		}
		try
		{
			var screenshots = new Regex("<div[^>]+id=\"thumbnails\"[^>]*>.*?</div>", RegexCompileFlags.DOTALL);
			desc = screenshots.replace(desc, -1, 0, "");
		}
		catch(RegexError e){
			critical("Failed to remove screenshots: %s", e.message);
		}
		
		try{
			var dis = new DataInputStream(template.read());
			string line;
			while ((line = dis.read_line (null)) != null){
				line = line.replace("${service.description}", desc);
				line = line.replace("${service.name}", service.name);
				line = line.replace("${service.version}", version);
				line = line.replace("${service.api_version}", api_version);
				line = line.replace("${service.maintainer_name}", maintainer_name);
				line = line.replace("${service.maintainer_link}", maintainer_link);
				line = line.replace("${service.resources}", resources);
				line = line.replace("${service.path}", path);
				line = line.replace("${service.requirements}", requirements);
				line = line.replace("${service.flash_enabled}", service.flash_plugin ? "Yes" : "No");
				builder.append(line);
				builder.append_c('\n');
			}
			dis.close();
		}
		catch(Error e){
			debug("ConfigFile: Unable to read %s", template.get_uri());
		}
		var buffer = page.buffer as Diorite.SimpleDocBuffer;
		try
		{
			buffer.load(builder.str);
		}
		catch(MarkupError e)
		{
			/// Error message for malformed service description 
			buffer.insert_at_cursor("\n\n" + _("Service description contains syntax error and might not be displayed correctly."), -1);
			warning("Failed to parse service description page : %s", e.message);
		}
	}
	
	private void on_add_service_clicked(){
		var dialog = new Gtk.FileChooserDialog(_("Open service integration package"),
			get_toplevel() as Gtk.Window, Gtk.FileChooserAction.OPEN, Gtk.Stock.CANCEL,
			Gtk.ResponseType.CANCEL, Gtk.Stock.OPEN, Gtk.ResponseType.ACCEPT);
		dialog.set_default_size(400, -1);
		var box = dialog.get_content_area() as Gtk.Box;
		if(box != null){
			var label = new Diorite.WrappedLabel(
			"New or updated service integrations are available at <a href='%s'>Nuvola Player Project website</a>.".printf(Config.UPDATE_PACKS_URL)
			+ " " + _("You can follow Nuvola Player on <a href='%1$s'>Google+</a>, <a href='%2$s'>Facebook</a> or <a href='%3$s'>Twitter</a> to be informed about new or updated packages.").printf(Config.GOOGLE_PLUS_URL, Config.FACEBOOK_URL, Config.TWITTER_URL), true);
			box.pack_start(label, false, false, 5);
			box.reorder_child(label, 0);
			label.show();
		}
		var response = dialog.run();
		var file = dialog.get_file();
		dialog.destroy();
		
		if(response == Gtk.ResponseType.ACCEPT){
			try{
				var service = manager.install_service(file);
				var info = new Diorite.Widgets.InfoDialog(_("Installation successfull"),
					_("Service %1$s (version %2$d.%3$d) has been installed succesfuly").printf(service.name, service.version, service.version_minor));
				
				reload(service.id);
				info.run();
			}
			catch(ServiceError e){
				var error = new Diorite.Widgets.ErrorDialog(_("Installation failed"),
					_("Installation of service from package %s failed.").printf(file.get_path())
					+ "\n\n" + e.message);
				error.run();
			}
		}
	}
	
	private void on_remove_service_clicked(){
		var service_id = service.id;
		try{
			manager.remove_service(service);
			var info = new Diorite.Widgets.InfoDialog(_("Removal successfull"),
				_("Service %1$s (version %2$d.%3$d) has been succesfuly removed").printf(service.name, service.version, service.version_minor));
			service = null;
			reload(service_id);
			info.run();
		}
		catch(ServiceError e){
			var error = new Diorite.Widgets.ErrorDialog(_("Removal failed"),
				_("Removal of service %s failed.").printf(service.name)
				+ "\n\n" + e.message);
			service = null;
			reload(service_id);
			error.run();
		}
	}
	
	private enum SMI{
		ID,
		ICON,
		NAME,
		COUNT;
	}
}

} // namespace Nuvola
