/*
 * Copyright 2013 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

namespace Nuvola
{

public class ServiceUpdatesDialog: Gtk.Dialog
{
	private ServiceUpdates updates;
	private Gtk.Grid grid;
	
	/**
	 * Creates new Service updates dialog
	 */
	public ServiceUpdatesDialog(Diorite.Application app, ServiceUpdates updates)
	{
		Object(window_position: Gtk.WindowPosition.CENTER);
		set_default_size(450, 450);
		/// Title of service selector window, %s will be replaced by application name
		title = _("Service updates - %s").printf(app.display_name);
		try
		{
			icon = Gtk.IconTheme.get_default().load_icon(app.icon, 48, 0);
		}
		catch(Error e)
		{
			warning("Unable to load application icon.");
		}
		this.updates = updates;
		
		var main_box = new Gtk.Grid();
		main_box.orientation = Gtk.Orientation.VERTICAL;
		var view = new Gtk.ScrolledWindow(null, null);
		grid = new Gtk.Grid();
		grid.orientation = Gtk.Orientation.VERTICAL;
		grid.show();
		view.add_with_viewport(grid);
		view.vexpand = true;
		view.show();
		main_box.add(view);
		get_content_area().add(main_box);
		
		var button = new Gtk.Button.from_stock(Gtk.Stock.REFRESH);
		button.clicked.connect(refresh);
		(get_action_area() as Gtk.Container).add(button);
		add_button(Gtk.Stock.CLOSE, Gtk.ResponseType.CLOSE);
		add_content();
	}
	
	public override bool delete_event(Gdk.EventAny event)
	{
		return true;
	}
	
	private void refresh()
	{
		updates.check_updates.begin(refresh_callback);
	}
	
	private void refresh_callback(GLib.Object? object, AsyncResult res)
	{
		int new_services;
		int new_updates;
		try
		{
			updates.check_updates.end(res, out new_services, out new_updates);
			foreach(var child in grid.get_children())
			{
				var item = child as UpdateItem;
				if (item != null)
					item.install.disconnect(on_install_update);
				grid.remove(child);
			}
			add_content();
		}
		catch (GLib.Error e)
		{
			var error = new Diorite.Widgets.ErrorDialog(_("Failed to get information about updates"),
				e.message);
			if (e is ServiceError.SERVER_ERROR_MESSAGE)
				error.secondary_use_markup = true;
			error.run();
		}
	}
	
	private void add_content()
	{
		SList<ServiceUpdate> new_services;
		SList<ServiceUpdate> new_updates;
		updates.get_updates(out new_services, out new_updates);
		if (new_services.length() > 0 || new_updates.length() > 0)
		{
			append_updates(new_services);
			append_updates(new_updates);
		}
		else
		{
			var label = new Gtk.Label(_("No update is available."));
			label.hexpand = true;
			label.vexpand = true;
			grid.add(label);
		}
		show_all();
	}
	
	private void append_updates(SList<ServiceUpdate> updates)
	{
		foreach (var update in updates)
		{
			var item = new UpdateItem(update);
			item.install.connect(on_install_update);
			grid.add(item);
		}
	}
	
	private void on_install_update(UpdateItem item)
	{
		try
		{
			var service = updates.install_update(item.update);
			item.install.disconnect(on_install_update);
			grid.remove(item);
			var info = new Diorite.Widgets.InfoDialog(_("Installation successfull"),
				_("Service %1$s (version %2$d.%3$d) has been installed succesfuly").printf(service.name, service.version, service.version_minor));
			info.run();
		}
		catch (ServiceError e)
		{
			var error = new Diorite.Widgets.ErrorDialog(_("Installation failed"),
				_("Installation of service failed.") + "\n\n" + e.message);
			error.run();
		}
	}
	
	private class UpdateItem: Gtk.Frame
	{
		public ServiceUpdate update {get; private set;}
		private Gtk.Grid? grid;
		private Gtk.Button button;
		
		public UpdateItem(ServiceUpdate update)
		{
			Object(shadow_type: Gtk.ShadowType.OUT);
			hexpand = true;
			this.update = update;
			grid = new Gtk.Grid();
			grid.margin = 5;
			
			var header = new Gtk.Label("<b>%s %d.%d</b> (%s)".printf(update.name, update.version_major, update.version_minor, update.is_new ? _("new service") : _("new version")));
			header.use_markup = true;
			header.hexpand = true;
			header.halign = Gtk.Align.START;
			grid.attach(header, 0, 0, 1, 1);
			
			button = new Gtk.Button.with_label(_("Install"));
			button.clicked.connect(on_button_clicked);
			grid.attach(button, 1, 0, 1, 1);
			
			var changes = new Gtk.Label(update.changes);
			changes.hexpand = true;
			changes.halign = Gtk.Align.START;
			grid.attach(changes, 0, 1, 2, 1);
			
			add(grid);
			grid.show_all();
		}
		
		public signal void install();
		
		private void on_button_clicked()
		{
			button.sensitive = false;
			button.label = _("Installing");
			install();
			button.label = _("Install");
			button.sensitive = true;
		}
	}
}

} // namespace Nuvola
