/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Diorite;
using Diorite.Logger;
using Diorite.System;
using Nuvola.Settings;
using Nuvola.Actions;
using Config;

[CCode(cheader_filename = "config.h")]
namespace Config
{
	[CCode(cname="G_LOG_DOMAIN")]
	extern extern const string LOG_DOMAIN;
	[CCode(cname="GETTEXT_PACKAGE")]
	extern const string GETTEXT_PACKAGE;
	[CCode(cname="LOCALES_DIR")]
	extern const string LOCALES_DIR;
	extern const string APPNAME;
	extern const string DESKTOP_ENTRY;
	extern const string UI_STATE_FILE;
	extern const string UNIQUE_NAME;
	extern const string APP_ICON;
	extern const string DISPLAY_NAME;
	extern const string VERSION;
	extern const string SERIES;
	extern const string SERIES_URL;
	extern const string SUPPORT_URL;
	extern const string BUG_TRACKER;
	extern const string REVISION;
	extern const string REVISION_ID;
	extern const string CONFIG_FILE;
	extern const string COOKIES_FILE;
	extern const string DEV_STAGE;
	extern const string PREFERENCES_HELP_URL;
	extern const string UPDATE_PACKS_URL;
	extern const string DONATE_URL;
	extern const string SCREENSHOTS_URL;
	extern const string ENABLED_FEATURES;
	extern const string DISABLED_FEATURES;
}

namespace Nuvola{

private const string APP_CRASHED = "crashed";
private const string APP_CHECK_CRASHED = "check_crashed";
private const string LAST_SERVICE_KEY = "last_service";

/**
 * Struct containing command line arguments. Check Args.options for their meaning.
 */
struct Args {
	static bool debug;
	static bool verbose;
	static bool version;
	static bool time;
	static bool debug_javascript;
	static bool debug_navigation;
	static bool debug_http;
	static bool no_web_plugins;
	static string? services_dir = null;
	
	public static const OptionEntry[] options = {
		{ "version", 'V', 0, GLib.OptionArg.NONE, ref Args.version, "Show version of program", null },
		{ "verbose", 'v', 0, GLib.OptionArg.NONE, ref Args.verbose, "Print information to terminal", null },
		{ "debug", 'D', 0, GLib.OptionArg.NONE, ref Args.debug, "Print debug messages", null },
		{ "debug-javascript", 0, 0, GLib.OptionArg.NONE, ref Args.debug_javascript, "Enable JavaScript debugging.", null },
		{ "debug-navigation", 0, 0, GLib.OptionArg.NONE, ref Args.debug_navigation, "Enable navigation debugging.", null },
		{ "debug-http", 0, 0, GLib.OptionArg.NONE, ref Args.debug_http, "Enable debugging of HTTP communication.", null },
		{ "time", 'T', 0, GLib.OptionArg.NONE, ref Args.time, "Show time in debug messages", null },
		{ "no-web-plugins", 0, 0, GLib.OptionArg.NONE, ref Args.no_web_plugins, "Disables web plugins.", null },
		{ "services-dir", 'S', 0, GLib.OptionArg.FILENAME, ref Args.services_dir, "Search for service integrations only in directory DIR and disable service management.", "DIR" },
		{ null }
	};
}

/**
 * Main entry point
 */
int main(string[] args){
	var start_time = TimeVal();
	start_time.get_current_time();
	
	/* 
	 * Work-around to get rid of "The following plugin is required: ABOUT protocol source" dialog.
	 * https://bugs.launchpad.net/nuvola-player/+bug/1320663
	 * https://bugs.webkit.org/show_bug.cgi?id=133605
	 */
	Environment.set_variable("GST_INSTALL_PLUGINS_HELPER", "/usr/bin/true", true);
	
	Gtk.init(ref args);
	// Save command and arguments
	var command = string.joinv(" ", args);
	
	/* Parse command line arguments */
	var opt_context = new OptionContext("");
	try{
		opt_context.set_help_enabled(true);
		opt_context.add_main_entries(Args.options, null);
		opt_context.parse(ref args);
	}
	catch(OptionError e){
		/// Shown if unsupported command line argument is used
		stderr.printf(_("Error: Option parsing failed:") + " %s\n\n%s\n",
		e.message, opt_context.get_help(true, null));
	return 1;
	}
	
	// Show version and exit
	if(Args.version){
		stdout.printf("%s %s\n", DISPLAY_NAME, VERSION);
		stdout.printf("Revision: %s, %s\n", REVISION, REVISION_ID);
		return 0;
	}
	
	/* Logger initialization.
	 * Use Diorite.debug() instead of GLib.debug() etc. to use "Nuvola" domain.
	 */
	Diorite.Logger.init("Nuvola", Args.debug ? LogLevelFlags.LEVEL_DEBUG
		: (Args.verbose ? LogLevelFlags.LEVEL_INFO : LogLevelFlags.LEVEL_WARNING));
	if(Args.time){
		Diorite.Logger.time = start_time.tv_sec;
	}
	
	Environment.set_variable("XLIB_SKIP_ARGB_VISUALS", "1", false);
	
	#if STABLE
		message("Welcome to %s, version: %s", DISPLAY_NAME, VERSION);
	#else
		message("Welcome to %s, development version: %s", DISPLAY_NAME, VERSION);
	#endif
	message("Revision: %s, %s", REVISION, REVISION_ID);
	message("Report any issues/bugs you might find to %s", BUG_TRACKER);
	debug("command: %s", command);
	
	debug("Enabled features: %s", ENABLED_FEATURES);
	debug("Disabled features: %s", DISABLED_FEATURES);
	
	/* Init translations */
	Intl.textdomain(APPNAME);
	Intl.bindtextdomain(APPNAME, LOCALES_DIR);
	
	var app = new Main();
	return app.run();
}

private class Main: Diorite.Application {
	private MainWindow main_window = null;
	private MainWebView main_view;
	private WebPlayer player;
	private Service? service = null;
	private Connection connection;
	private ServicesManager services_manager;
	private ObjectContainer objects;
	private Configuration config;
	private MapFile ui_state = null;
	private ExtensionsManager extensions;
	private Storage storage;
	private ServiceSelectorDialog? selector = null;
	private Diorite.Actions actions;
	private Diorite.ActionGroups action_groups;
	private Gtk.UIManager ui_manager;
	private PreferencesWindow? preferences_window = null;
	private bool loaded = false;
	private NoServiceOverlay? no_service_overlay = null;
	private WebBackend web_backend = null;
	private FormatSupport formats;
	private Core core;
	private DownloadsDialog? downloads = null;
	private UpdatesService updates;

	/**
	 * Constructor for the main application class
	 * 
	 * @param config Configuration
	 * @param ui_state User interface state
	 * @param storage File storage
	 */
	public Main()
	{
		base(UNIQUE_NAME, DESKTOP_ENTRY, APPNAME, null);
		icon = APP_ICON;
		version = VERSION;
		display_name = DISPLAY_NAME;
		app_name = APPNAME;
		
		activate.connect(activated);
	}
	
	#if DEBUG_MEMORY
	~Main(){
		debug("~Main");
	}
	#endif
	
	/**
	 * Called when application is about to be activated.
	 * Initializes new instance or activated already
	 * running instance.
	 */
	private void activated()
	{
		if (main_window == null)
		{
			shutdown.connect(on_shutdown);
			shutdown.connect_after(on_late_shutdown);
			start();
		}
		else{
			main_window.present();
		}
	}
	
	/**
	 * Start application
	 */
	private void start()
	{
		debug("Starting new instance");
		/*
		 * Set up core
		 */
		#if UNSTABLE
		storage = new XdgStorage.for_project(APPNAME, "_unstable");
		#else
		storage = new XdgStorage.for_project(APPNAME);
		#endif
		config = new Configuration(storage.get_config_path(CONFIG_FILE));
		var logging = Args.debug_http ? Soup.LoggerLogLevel.HEADERS : Soup.LoggerLogLevel.NONE;
		connection = new Connection.with_session(WebKit.get_default_session(), storage, logging);
		formats = new FormatSupport(storage.get_data_file("audio/audiotest.mp3").get_uri());
		core = new Core(this, storage, config, connection, formats);
		
		/*
		 * Set up UI actions and state
		 */
		ui_manager = new Gtk.UIManager();
		actions = new Diorite.Actions(ui_manager, app_name);
		action_groups = new Diorite.ActionGroups();
		ui_state = new MapFile(storage.get_config_path(UI_STATE_FILE));
		
		
		/*
		 * Set up services manager
		 */
		var services_storage = storage.get_child(SERVICES_DIRNAME);
		var allow_service_management = true;
		if (Args.services_dir != null && Args.services_dir != "")
		{
			var services_dir = File.new_for_path(Args.services_dir);
			if (!services_dir.query_exists())
			{
				warning("Services dir %s doesn't exist, ignoring.", Args.services_dir);
			}
			else
			{
				debug("Using services dir: %s", services_dir.get_path());
				services_storage = new Diorite.Storage(services_dir.get_path(), {},
				services_storage.user_config_dir.get_path(),
				services_storage.user_cache_dir.get_path());
				allow_service_management = false;
			}
		}
		services_manager = new ServicesManager(services_storage, allow_service_management);
		services_manager.service_installed.connect(on_service_installed_or_removed);
		services_manager.service_removed.connect(on_service_installed_or_removed);
		
		
		/*
		 * Set up web backend
		 */
		main_view = new MainWebView(storage.get_child("webkit"), actions);
		web_backend = new WebBackend(this, storage, main_view, config, ui_state, formats, actions);
		web_backend.web_plugins = !Args.no_web_plugins;
		web_backend.web_plugins_blocker = config.web_plugins_blocker;
		web_backend.download_started.connect(on_download_started);
		player = new WebPlayer(web_backend, config, storage, connection, actions, action_groups);
		config.value_changed.connect(on_settings_changed);
		player.song_changed.connect(on_song_changed);
		
		/*
		 * Start ui
		 */
		show_user_info.connect(show_info);
		append_ui_actions();
		prepare_gui();
		load_extensions();
		loaded = true;
		start_or_select_service();
		
		var updates = new ServiceUpdates(services_manager, connection,
		config.updates_server,
		"%d.%d".printf(JSApi.API_VERSION, JSApi.API_VERSION_MINOR));
		this.updates = new UpdatesService(this, config, updates, main_window);
	}
	
	/**
	 * Show information to user
	 * 
	 * @param title    title of a message
	 * @param text     text of a message
	 */
	public void show_info(string title, string text)
	{
		var dialog = new Diorite.Widgets.InfoDialog(title, text);
		dialog.run();
		while (Gtk.events_pending())
			Gtk.main_iteration();
	}
	
	/**
	 * Adds user interface actions related to main application object
	 */
	private void append_ui_actions(){
		actions.add_actions({
			/// Menu Application
			new NormalAction("application", null, _("_Application"), null, null),
			/// Quit action (show in menus)
			new NormalAction("quit", Gtk.Stock.QUIT, null, "<ctrl>Q", null, quit),
			/// Switch service action (show in menus)
			new NormalAction("switch-service", "nuvolaplayer", _("_Switch Service"), "<ctrl>S",
			/// tooltip of switch service action
			_("Switch service"), select_service),
			/// Menu Go
			new NormalAction("go", null, _("_Go"), null, null),
			/// Menu Control
			new NormalAction("control", null, _("_Control"), null, null),
			/// Menu View
			new NormalAction("view",null, _("_View"), null, null),
			/// Remove cookies action (shown in Preferences dialog, tab Advanced)
			new NormalAction(PURGE_COOKIES, null, _("Remove cookies"), null, null, core.purge_cookies),
			/// Purge cache action (shown in Preferences dialog, tab Advanced)
			new NormalAction(PURGE_CACHE, null, _("Remove files from cache"), null, null, purge_cache),
			/// Reset state of user interface action (shown in Preferences dialog, tab Advanced)
			new NormalAction(PURGE_UI_STATE, null, _("Reset state of user interface"), null, null, purge_ui_state),
			/// Purge settings action (shown in Preferences dialog, tab Advanced)
			new NormalAction(PURGE_SETTINGS, null, _("Reset settings"), null, null, core.purge_settings),
			/// Format support action (shown in menu)
			new NormalAction("format-support", null, _("Format support"), null, null, format_support),
			/// Preferences action (shown in menu)
			new NormalAction("preferences", Gtk.Stock.PREFERENCES, null, null, null, edit_settings),
			/// Service settings action
			new NormalAction("service-settings", null, _("Service settings"), null, null, edit_service),
			/// Extensions settings action
			new NormalAction("extensions-settings", Gtk.Stock.PREFERENCES, _("Extensions settings"), null, null, edit_extensions),
			/// Menu Help
			new NormalAction("help", null, _("_Help"), null, null),
			/// Support action (shown in menu)
			new NormalAction("questions", Gtk.Stock.HELP, _("Get help online"), null, null, open_support),
			new NormalAction("about", Gtk.Stock.ABOUT, null, null, null, about),
			/// Release notes action (shown in menu)
			new NormalAction("website", Gtk.Stock.INFO, _("Release notes"), "F1", null, open_release_notes),
			/// Open downloads action (shown in menu)
			new NormalAction("downloads", null, _("Downloads"), null, null, open_downloads),
			new NormalAction("updates", null, _("Service updates manager"), null, null, show_updates)
		});
	}
	
	/**
	 * Load main graphical user interface
	 */
	private void prepare_gui(){
		var settings = Gtk.Settings.get_default();
		settings.gtk_application_prefer_dark_theme = config.dark_theme;
		main_window = new MainWindow(this, ui_manager, actions, storage, ui_state);
		main_window.close_action.connect_after(on_close_action);
		
		if (Args.debug_javascript)
			main_view.developer_mode = true;
		if (Args.debug_navigation)
			main_view.log_navigation = true;
		main_view.populate_popup.connect(on_populate_popup);
		
		main_window.content.add(main_view);
		main_window.load_actions_ui();
		main_view.grab_focus();
		update_window_title();
		
		donation_bar();
		
		main_window.show_all();
	}
	
	/**
	 * Prepares resources for extensions and loads extensions.
	 */
	private void load_extensions(){
		var objects = new HashTable<string, Object>(str_hash, str_equal);
		objects.insert("application", this);
		objects.insert("main_window", main_window);
		objects.insert("main_view", main_view);
		objects.insert("player", player);
		objects.insert("actions", actions);
		objects.insert("action_groups", action_groups);
		objects.insert("storage", storage);
		objects.insert("config", config);
		objects.insert("ui_state", ui_state);
		objects.insert("services_manager", services_manager);
		objects.insert("web_backend", web_backend);
		this.objects = new ObjectContainer(objects);
		
		extensions = new ExtensionsManager(ui_manager, main_window, this.objects);
		var available_extensions = extensions.available_extensions;
		foreach (var key in available_extensions.get_keys())
			if (config.get_bool("extension.%s.enabled".printf(key), available_extensions.lookup(key).autoload))
				extensions.load(key);
	}
	
	/**
	 * Starts last service or shows service selector
	 */
	private void start_or_select_service(){
		ui_state["release"] = SERIES;
		ui_state["dev_channel"] = DEV_STAGE;
		var service_id = ui_state[LAST_SERVICE_KEY];
		if (service_id != null)
		{
			var new_service = services_manager.get_service(service_id);
			if (new_service != null)
			{
				load_service(new_service);
				return;
			}
		}
		no_service_overlay = new NoServiceOverlay(actions);
		no_service_overlay.show();
		main_window.overlay.add_overlay(no_service_overlay);
	}
	
	private void reload_service(string service_id)
	{
		var new_service = services_manager.get_service(service_id);
		if (new_service != null)
		{
			load_service(new_service);
		}
		else
		{
			if (player != null)
				player.pause();
			web_backend.load_service(null);
			main_view.service = null;
			service = null;
			main_view.open("about:blank");
			no_service_overlay = new NoServiceOverlay(actions);
			no_service_overlay.show();
			main_window.overlay.add_overlay(no_service_overlay);
		}
	}
	/**
	 * Loads service
	 */
	private void load_service(Service new_service){
		this.service = new_service;
		web_backend.load_service(new_service);
	}
	
	/**
	 * Show service selector
	 */
	public void select_service(){
		if (selector != null)
		{
			if (service != null)
			{
				destroy_selector();
			}
			else
			{
				selector.present();
			}
			return;
		}
		
		if (preferences_window != null)
		{
			// Preferences contains service-specific settings!
			preferences_window.close();
		}
		
		string? current_service = service != null ? service.id : null;
		debug("Select service, prev = %s", current_service ?? "(none)");
		var template = storage.get_data_file("html/Service.html");
		var view = new ServicesManagerView(icon, template, services_manager,
			current_service);
		selector = new ServiceSelectorDialog(this, view);
		view.response.connect(on_selector_response);
		selector.set_transient_for(main_window);
		selector.show_all();
	}
	
	/**
	 * Called when service is selected
	 */
	private void on_selector_response(int response_id)
	{
		var new_service = selector.view.service;
		destroy_selector();
		if (response_id == Gtk.ResponseType.OK
		&& (service == null || service.id != new_service.id))
		{
			if (service == null && no_service_overlay != null)
			{
				main_window.overlay.remove(no_service_overlay);
				no_service_overlay.destroy();
				no_service_overlay = null;
			}
			
			load_service(new_service);
			ui_state[LAST_SERVICE_KEY] = new_service.id;
		}
	}
	
	/**
	 * Destroys the Service selector screen
	 */
	private void destroy_selector(){
		return_if_fail(selector != null);
		selector.view.response.disconnect(on_selector_response);
		selector.destroy();
		selector = null;
	}
	
	/**
	 * Opens About dialog.
	 */
	private void about(){
		var dialog = new AboutDialog(this);
		dialog.run();
		dialog.destroy();
	}
	
	/**
	 * Opens Preferences dialog with General tab selected.
	 */
	private void edit_settings(){
		show_preferences();
	}
	
	/**
	 * Opens Preferences dialog with Service tab selected.
	 */
	private void edit_service(){
		show_preferences(PreferencesTab.SERVICE);
	}
	
	/**
	 * Opens Preferences dialog with Extensions tab selected.
	 */
	private void edit_extensions(){
		show_preferences(PreferencesTab.EXTENSIONS);
	}
	
	/**
	 * Opens Preferences dialog
	 * 
	 * @param tab	a tab to be initially visible
	 */
	private void show_preferences(PreferencesTab tab=PreferencesTab.GENERAL){
		if(preferences_window != null){
			preferences_window.present();
			return;
		}
		
		preferences_window = new PreferencesWindow(objects, web_backend, this, this.storage, this.connection, services_manager, extensions, tab);
		preferences_window.present();
		preferences_window.response.connect(on_preferences_window_response);
	}
	
	private void on_preferences_window_response(int response_id){
		if(preferences_window == null) return;
		if(response_id == Gtk.ResponseType.HELP){
			open_uri_with_fallback(PREFERENCES_HELP_URL);
		}
		else{
			preferences_window.close();
			preferences_window = null;
		}
	}
	
	/**
	 * Opens project support page
	 */
	private void open_support(){
		open_uri_with_fallback(SUPPORT_URL);
	}
	
	/**
	 * Opens release notes for current release
	 */
	private void open_release_notes(){
		open_uri_with_fallback(SERIES_URL);
	}
	
	/**
	 * Called when application is about to be terminated
	 */
	public void on_shutdown(){
		debug("Early shutdown");
		if(player != null){
			player.pause();
		}
		
		if (preferences_window != null){
			preferences_window.close();
			preferences_window = null;
		}
		
		if(main_window != null && main_window.visible){
			this.main_window.hide();
		}
		
		extensions.unload_all();
	}
	
	/**
	 * Called when application is about to be terminated,
	 * but after all early callbacks
	 */
	public void on_late_shutdown(){
		return_if_fail(ui_state != null);
		debug("Late shutdown");
		ui_state.set_bool(APP_CRASHED, false);
		try
		{
			ui_state.save();
			debug("User interface state saved.");
		}
		catch (Error e)
		{
			warning("Unable to save UI state: %s", e.message);
		}
	}
	
	/**
	 * Called when the main window is about to be closed.
	 */
	private void on_close_action()
	{
		var hide = config.hide_window;
		if(hide == OptionState.OFF || hide == OptionState.AUTO && player.playback_state != STATE_PLAYING)
		{
			debug("On close: quit");
			quit();
		}
		else
		{
			debug("On close: hide");
		}
	}
	
	private void on_donate_button_clicked()
	{
		Diorite.System.open_uri_with_fallback(Config.DONATE_URL);
	}
	
	private void on_service_installed_or_removed(string service_id)
	{
		if (service != null && service.id == service_id)
			reload_service(service_id);
	}
	
	/**
	 * Purge user interface state and show request to restart application.
	 */
	private void purge_ui_state(){
		try{
			ui_state.purge();
		}
		catch(Error e){
			critical("Unable to purge UI state: %s", e.message);
		}
		
		if(loaded){
			main_view.go_home();
			//// Title in information dialog when default settings were restored
			var dialog = new Diorite.Widgets.InfoDialog(_("Default settings were restored."),
				/// Request to restart application
				_("Please restart application to finish the process."));
			dialog.run();
		}
	}
	
	private void donation_bar()
	{
		Soup.Message message;
		message = new Soup.Message("GET", "https://www.gittip.com/NuvolaPlayer/public.json");
		var session = WebKit.get_default_session();
		session.queue_message(message, donation_bar_callback);
	}
	
	private void donation_bar_callback(Soup.Session session, Soup.Message message)
	{
		var response = (string) message.response_body.flatten().data;
		var parser = new Json.Parser();
		try
		{
			parser.load_from_data(response);
		}
		catch (GLib.Error e)
		{
			warning("Unexpected response from Gittip: invalid JSON document. %s", e.message);
			return;
		}
		
		var root_node = parser.get_root();
		if (root_node == null || root_node.get_node_type() != Json.NodeType.OBJECT)
		{
			warning("Unexpected response from Gittip: invalid JSON document");
			return;
		}
		
		var root_object = root_node.get_object();
		
		string goal = null;
		Json.Node member;
		if (root_object.has_member("goal"))
		{
			member = root_object.get_member("goal");
			if (member.get_node_type() == Json.NodeType.VALUE
			&& member.get_value_type() == typeof(string))
			{
				goal = member.dup_string();
			}
		}
		
		string receiving = null;
		if (root_object.has_member("receiving"))
		{
			member = root_object.get_member("receiving");
			if (member.get_node_type() == Json.NodeType.VALUE
			&& member.get_value_type() == typeof(string))
			{
				receiving = member.dup_string();
			}
		}
		
		string text = null;
		if (receiving != null && goal != null)
		{
			var percentage = (int)(100*double.parse(receiving)/double.parse(goal));
			if (percentage < 100)
			{
				text = "<b>Nuvola Player is currently receiving $%s of the $%s weekly funding goal.</b>\nPlease donate to help keep Nuvola Player going.".printf(receiving, goal);
				var donate_bar = new Diorite.InfoBar.info(text);
				var donate_button = new Gtk.Button.with_label(_("Make a donation"));
				donate_button.clicked.connect(on_donate_button_clicked);
				donate_bar.add_button(donate_button);
				main_window.box.pack_end(donate_bar, false);
				donate_bar.show();
			}
		}
	}
	
	private void purge_cache()
	{
		do_purge_cache.begin();
	}
	
	private async void do_purge_cache()
	{
		var dialog = (Gtk.MessageDialog) Object.new(typeof(Gtk.MessageDialog),
				title: "",
				modal: true,
				message_type: Gtk.MessageType.INFO
			);
		/// Action label
		dialog.text = _("Remove files from cache");
		/// Operation in progress
		dialog.secondary_text = _("Operation is in progress. Please wait.");
		dialog.show_all();
		while (Gtk.events_pending())
			Gtk.main_iteration();
		Idle.add(do_purge_cache.callback);
		yield;
		yield core.purge_cache();
		dialog.destroy();
		show_user_info(/// Title in information dialog
			_("Cache has been purged."),
			/// Request to reload page after cache being purged
			_("You can reload the page to get fresh data."));
	}
	
	/**
	 * Called when settings have been changed.
	 */
	private void on_settings_changed(string key){
		switch(key){
		case Settings.WEB_PLUGINS_BLOCKER:
			web_backend.web_plugins_blocker = config.web_plugins_blocker;
			break;
		case Settings.DARK_THEME:
				var settings = Gtk.Settings.get_default();
				settings.gtk_application_prefer_dark_theme = config.dark_theme;
			break;
		}
	}
	
	private void on_song_changed(string? song, string? artist, string? album, string? album_art){
		update_window_title();
	}
	
	private void format_support()
	{
		var dialog = new FormatSupportDialog(formats.flash_available, formats.mp3_available, storage, main_window);
		dialog.present();
	}
	
	private void update_window_title(){
		if(selector != null){
			/// Title of service selector window
			main_window.title = _("Select service") + " - " + display_name;
		}
		else if(service == null){
			main_window.title = display_name;
		}
		else if(player.playback_state == STATE_NONE || player.song == null){
			main_window.title = "%s - %s".printf(service.name, display_name);
		}
		else if(player.artist == null){
			main_window.title = "%s - %s - %s".printf(player.song, service.name, display_name);
		}
		else{
			main_window.title = "%s - %s - %s - %s".printf(player.song, player.artist, service.name, display_name);
		}
	}
	
	private void open_downloads()
	{
		if (downloads == null)
			downloads = new DownloadsDialog(this);
		downloads.present();
	}
	
	private void show_updates()
	{
		updates.show_updates();
	}
	
	/**
	 * Modifies popup menu of WebKit WebView
	 */
	private void on_populate_popup(WebKit.WebView web_view, Gtk.Menu menu)
	{
		var children = menu.get_children();
		var size = children.length();
		
		if(Args.debug_javascript && size >= 2){
			/* Inspector item is the last menu item and is preceded by
			 * a separator */
			var inspector_item = children.nth_data(size - 1) as Gtk.MenuItem;
			var separator_item = children.nth_data(size - 2) as Gtk.SeparatorMenuItem;
			if (separator_item != null && inspector_item != null)
			{
				size -= 2; // Don't remove the last two items
				menu.remove(separator_item);
				menu.append(separator_item);
				separator_item.show();
			}
		}
		
		for(var i =0; i < size; i++){
			menu.remove(children.nth_data(i));
		}
		
		main_window.populate_popup_menu(menu);
	}
	
	private void on_download_started(WebKit.Download request, File destination)
	{
		open_downloads();
		downloads.add_download(request, destination);
	}
}

} // namespace Nuvola
